import pytest
import os
import locidex
#from locidex.classes.blast import blast_search, parse_blast
from locidex.classes.blast import FilterOptions, BlastSearch, BlastMakeDB
from locidex.constants import BlastColumns, BlastCommands
from locidex.classes.extractor import extractor
from locidex.classes.seq_intake import seq_intake
from locidex.classes.db import db_config
from locidex.manifest import DBData
from pathlib import Path

#could be tested via locidex extract -i ./locidex/example/search/NC_003198.1.fasta -d ./locidex/example/build_db_mlst_out/  -o tmp --force

PACKAGE_ROOT = os.path.dirname(locidex.__file__)



def blast_db_and_search(tmpdir,input_db_path):
    #blast_search_obj = blast_search(input_db_path=input_db_path, 
    #                    input_query_path=os.path.join(PACKAGE_ROOT, 'example/build_db_mlst_out/blast/nucleotide/nucleotide.fasta'),
    #                    output_results=os.path.join(tmpdir,"hsps.txt"), blast_params={'evalue': 0.0001,'max_target_seqs': 10,'num_threads': 1}, 
    #                    blast_method='blastn',
    #                    blast_columns=BlastColumns._fields,create_db=True)
    #blast_search_obj.run_blast()
    #output_blast_results_path = os.path.join(tmpdir,"hsps.txt")
    #parse_blast_obj = parse_blast(input_file = output_blast_results_path,
    #                                                    blast_columns = BlastColumns._fields,
    #                                                    filter_options={'bitscore':{'min':600, 'max':None, 'include':None}})

    blast_search_obj = BlastMakeDB(
        db_data=Path(input_db_path),
        db_type=DBData.nucleotide_db_type(),
        parse_seqids=True,
        output_db_path=Path(input_db_path))
    blast_db_path = blast_search_obj.makeblastdb()

    blast_search_obj = BlastSearch(db_data=blast_db_path, 
                query_path=Path(os.path.join(PACKAGE_ROOT, 'example/build_db_mlst_out/blast/nucleotide/nucleotide.fasta')),
                blast_params={'evalue': 0.0001,'max_target_seqs': 10,'num_threads': 1},
                blast_method=BlastCommands.blastn,
                blast_columns=BlastColumns._fields,
                filter_options={'bitscore':FilterOptions(**{'min':600, 'max':None, 'include':None})})
    
    parse_blast_obj = blast_search_obj.get_blast_data(db_path=blast_db_path, output=tmpdir.joinpath("hsps.txt"))
    
    #return parse_blast_obj

@pytest.fixture
def seq_intake_fixture():
    # Mimicking the creation of seq_data from a given input fasta file
    input_fasta = Path(PACKAGE_ROOT).joinpath('example/search/NC_003198.1.fasta')
    format = "fasta" # Adjust this based on your file type
    translation_table = 11
    seq_obj = seq_intake(input_fasta, format, 'source', translation_table, perform_annotation=False,skip_trans=True)
    return seq_obj


def test_extractor_initialization(seq_intake_fixture, tmpdir):
    db_path=os.path.join(tmpdir,"contigs.fasta")
    nt_db_test = os.path.join(PACKAGE_ROOT,'example/build_db_mlst_out/blast/nucleotide/nucleotide.fasta')
    hit_file = os.path.join(tmpdir,"hsps.txt")
    blast_params={'evalue': 0.0001, 'max_target_seqs': 10, 'num_threads': 1}
    metadata_path = os.path.join(PACKAGE_ROOT,'example/build_db_mlst_out/meta.json')
    seq_obj = seq_intake_fixture
    seq_data={}
    with open(db_path,'w') as oh:
        for idx,seq in enumerate(seq_obj.seq_data):
            seq_data[str(idx)] = {'id':str(seq.seq_id),'seq':seq.dna_seq}
            oh.write(">{}\n{}\n".format(idx,seq.dna_seq))

    blast_db = BlastMakeDB(db_path, DBData.nucleotide_db_type(), True, db_path)
    blast_db.makeblastdb()
    blast_search_obj = BlastSearch(db_data=blast_db.output_db_path,
                                query_path=nt_db_test,
                                blast_params=blast_params,
                                blast_method=BlastCommands.blastn,
                                blast_columns=BlastColumns._fields)
    
    hit_df = blast_search_obj.get_blast_data(blast_db.output_db_path, Path(hit_file))
    loci = []; metadata_obj = db_config(metadata_path, ['meta', 'info'])
    for idx,row in hit_df.iterrows():
        qid = str(row['qseqid'])
        loci.append(metadata_obj.config['meta'][qid]['locus_name'])
    hit_df['locus_name'] = loci

    extractor_instance = extractor(
        df=hit_df,
        seq_data=seq_data,
        sseqid_col='sseqid',
        queryid_col='qseqid',
        qstart_col='qstart',
        qend_col='qend',
        qlen_col='qlen',
        sstart_col='sstart',
        send_col='send',
        slen_col='slen',
        sstrand_col='sstrand',
        bitscore_col='bitscore',
        extend_threshold_ratio = 0.2,
        filter_contig_breaks=True
    )
    extractor_instance.df.to_csv(os.path.join(tmpdir,'filtered.hsps.txt'),header=True,sep="\t",index=False)  

    # Verify modifications by the class's methods
    assert not extractor_instance.df.empty, "The extractor DataFrame should not be empty after initialization."

    # Assert modifications are correctly applied
    for col in ['qstart', 'qend', 'sstart', 'send']:
        assert all(extractor_instance.df[col] >= 0), f"Column {col} should have been adjusted for zero indexing."

    # Assert sequences are correctly extracted and populated in `seqs`
    assert extractor_instance.seqs, "Sequences should be extracted and the `seqs` dictionary should be populated."

    # Assert filtering based on contig boundaries if applicable
    if extractor_instance.filter_contig_breaks:
        assert not extractor_instance.df.query("`is_5prime_boundary` == False and `is_3prime_boundary`== False").empty, "Hits on contig boundaries should be filtered out when `filter_contig_breaks` is True with both 5' and 3' boundaries set to True."

    # Assert all expected columns are present after adjustments
    expected_columns = ['ext_start', 'ext_end', 'is_complete', 'is_5prime_boundary', 'is_3prime_boundary', 'reverse', 'complement', 'is_extended']
    for col in expected_columns:
        assert col in extractor_instance.df.columns, f"Expected column {col} to be present in the DataFrame."




import gzip
import json
import os
import pathlib
import sys
from argparse import ArgumentParser
from datetime import datetime
from functools import partial
from mimetypes import guess_type
from dataclasses import dataclass
from typing import List, Tuple
import logging
import errno
import pandas as pd
from Bio import SeqIO
from Bio.Seq import Seq
from pyrodigal import GeneFinder

from locidex.constants import FILE_TYPES, LocidexDBHeader, CharacterConstants, raise_file_not_found_e
from locidex.utils import six_frame_translation, revcomp, calc_md5
from locidex.version import __version__

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.INFO)

class locidex_format:

    input_type = None
    delim = '_'
    status = True
    __stop_codon = CharacterConstants.stop_codon

    # ? These two parameters below can probably be cleaned up
    __file_input = "file"
    __dir_input = "dir"

    @dataclass
    class FrameSelection:
        offset: int
        seq: str
        count_int_stops: int
        revcomp: bool

    def __init__(self,input,header,is_protein=False,delim="_",trans_table=11,
                min_len_frac=0.7,max_len_frac=1.3,min_cov_perc=80.0,min_ident_perc=80.0,valid_ext=None):
        self.input = input
        self.seq_idx = 0
        self.gene_name = None
        self.header = header
        self.delim = delim
        self.translation_table = trans_table
        self.min_len_frac = min_len_frac
        self.max_len_frac = max_len_frac
        self.min_cov_perc = min_cov_perc
        self.min_ident_perc = min_ident_perc
        self.is_protein_coding = is_protein
        self.data = dict()
        self.valid_ext = valid_ext

        if valid_ext is not None:
            if isinstance(valid_ext,list):
                self.valid_ext = valid_ext
            else:
                self.valid_ext = [valid_ext]

        self.set_input_type()
        if self.input_type == self.__dir_input:
            self.process_dir()
        else:
            self.parse_fasta(self.input)

    def process_dir(self) -> None:
        files = self.get_dir_files(self.input)
        for f in files[self.__file_input]:
            for e in self.valid_ext:
                if f[1].endswith(e):
                    self.gene_name = f[1].replace(f'{e}','')
                    self.parse_fasta(f[0])
                    break
            else:
                logger.critical("File: {} does not have a valid extension. Valid extensions are: {}".format(e, str(self.valid_ext)))
                raise ValueError("Extension for file: {} is not allowed.".format(e))

    def set_input_type(self):
        if os.path.isfile(self.input):
            self.input_type = self.__file_input
        elif os.path.isdir(self.input):
            self.input_type = self.__dir_input
        else:
            logger.critical("Could not determine input type for: {}".format(self.input))
            raise AttributeError("Unknown input type could not be determined for: {}".format(self.input))

    def get_dir_files(self, input_dir):
        files = {self.__file_input: [], self.__dir_input: []}
        d = pathlib.Path(input_dir)
        for item in d.iterdir():
            type = self.__file_input
            if item.is_dir():
                type = self.__dir_input

            files[type].append([f"{item.resolve()}", os.path.basename(item)])
        return files

    def pick_frame(self, six_frame_translation) -> FrameSelection:
        """
        Reducing the complexity of this function now only checking if the allele is reverse complimented
        """

        reversed_frame_idx = 3 # all frames above this index are reverse complimented
        fwd_idx, rev_idx = 0, 3
        reverse_p = False

        fwd = six_frame_translation[fwd_idx] # frame 1
        rev = six_frame_translation[rev_idx] # frame 2

        fwd_stop_counts = fwd[:-1].count(self.__stop_codon)

        output_seq = fwd
        idx = fwd_idx
        min_int_stop = fwd_stop_counts

        if rev[-1] == self.__stop_codon and (stop_counts := rev[:-1].count(self.__stop_codon)) == 1 and fwd_stop_counts > 0:
            idx = rev_idx
            output_seq = rev
            min_int_stop = stop_counts

        offset = idx % reversed_frame_idx # gives offset for both revcomp and seq
        if idx >= reversed_frame_idx:
            reverse_p = True

        return self.FrameSelection(offset=offset, seq=output_seq.lower(), count_int_stops=min_int_stop, revcomp=reverse_p)


    def parse_fasta(self, input_file):
        encoding = guess_type(input_file)
        _open = partial(gzip.open, mode='rt') if encoding == 'gzip' else open
        with _open(input_file) as f:
            for record in SeqIO.parse(f, 'fasta'):
                id_in = str(record.id)
                if self.input_type == self.__file_input:
                    gene_name = "_".join(id_in.split(self.delim)[:-1])
                else:
                    gene_name = self.gene_name
                dna_seq = str(record.seq).lower().replace('-','')
                aa_seq = ''
                if self.is_protein_coding:
                    t = self.pick_frame(six_frame_translation(dna_seq, trans_table=self.translation_table))
                    aa_seq = t.seq
                    dna_seq = dna_seq[t.offset:]
                    if t.revcomp:
                        dna_seq = revcomp(dna_seq)

                dna_len = len(dna_seq)
                aa_len = len(aa_seq)

                aa_encoding_p = lambda x: x if self.is_protein_coding else None
                row = LocidexDBHeader(
                    seq_id=self.seq_idx,
                    locus_name=gene_name,
                    locus_name_alt=id_in,
                    locus_product='',
                    locus_description='',
                    locus_uid=id_in.split(self.delim)[-1],
                    dna_seq=dna_seq,
                    dna_seq_len=dna_len,
                    dna_seq_hash=calc_md5([dna_seq])[0],
                    dna_ambig_count=dna_seq.count('n'),
                    aa_seq = aa_encoding_p(aa_seq),
                    aa_seq_len=aa_encoding_p(len(aa_seq)),
                    aa_seq_hash= aa_encoding_p(calc_md5([aa_seq])[0]),
                    aa_min_ident=aa_encoding_p(self.min_ident_perc * (self.min_ident_perc /100)),
                    aa_min_len=aa_encoding_p(aa_len * self.min_len_frac),
                    aa_max_len=aa_encoding_p( aa_len * self.max_len_frac),
                    min_aa_match_cov=aa_encoding_p(self.min_cov_perc),
                    count_int_stops=aa_encoding_p(t.count_int_stops),
                    dna_min_len=dna_len*self.min_len_frac,
                    dna_max_len=dna_len*self.max_len_frac,
                    dna_min_ident=self.min_ident_perc,
                    min_dna_match_cov=self.min_cov_perc
                )
                self.data[self.seq_idx] = row
                self.seq_idx += 1


def add_args(parser=None):
    if parser is None:
        parser = ArgumentParser(
            description="Locidex: Format sequences for a database.",)
    parser.add_argument('-i','--input', type=str, required=True,help='Input directory of fasta files or input fasta')
    parser.add_argument('-o', '--outdir', type=str, required=True, help='Output directory to put results')
    parser.add_argument('--min_len_frac', type=int, required=False, help='Used to calculate individual sequence minimum acceptable length (0 - 1)',
                        default=0.7)
    parser.add_argument('--max_len_frac', type=int, required=False, help='Used to calculate individual sequence maximimum acceptable length (1 - n)',
                        default=1.3)
    parser.add_argument('--min_ident', type=float, required=False, help='Global minumum percent identity required for match',
                        default=80.0)
    parser.add_argument('--min_match_cov', type=float, required=False, help='Global minumum percent hit coverage identity required for match',
                        default=80.0)
    parser.add_argument('--translation_table', type=int, required=False,
                        help='output directory', default=11)
    parser.add_argument('-n', '--not_coding', required=False, help='Skip translation',
                        action='store_true')
    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)
    parser.add_argument('-f', '--force', required=False, help='Overwrite existing directory',
                        action='store_true')
    return parser

def run(cmd_args=None):
    if cmd_args is None:
        parser = add_args()
        cmd_args = parser.parse_args()

    input = cmd_args.input
    outdir = cmd_args.outdir
    min_len_frac = cmd_args.min_len_frac
    max_len_frac = cmd_args.max_len_frac
    min_ident = cmd_args.min_ident
    min_match_cov = cmd_args.min_match_cov
    trans_table = cmd_args.translation_table
    force = cmd_args.force

    is_coding = True
    if cmd_args.not_coding:
        is_coding = False

    run_data = dict()
    run_data['analysis_start_time'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    run_data['parameters'] = vars(cmd_args)

    if os.path.isdir(outdir) and not force:
        logger.critical(f'Error {outdir} exists, if you would like to overwrite, then specify --force')
        raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST), str(outdir))


    if not os.path.isdir(outdir):
        os.makedirs(outdir, 0o755)

    if not os.path.isdir(input) and not os.path.isfile(input):
        logger.critical(f'Error {input} does not exist as a file or directory')
        raise_file_not_found_e(input, logger)
    
    logger.info("Beginning format operation.")
    obj = locidex_format(input=input,header=LocidexDBHeader._fields,is_protein=is_coding,min_len_frac=min_len_frac,max_len_frac=max_len_frac, min_ident_perc=min_ident,
            min_cov_perc=min_match_cov,trans_table=trans_table,valid_ext=list(FILE_TYPES['fasta']))
    logger.info("Finished format.")
    run_data['result_file'] = os.path.join(outdir,"locidex.txt")
    pd.DataFrame.from_dict(obj.data,orient='index').to_csv(run_data['result_file'],sep="\t",index=False,header=True)

    run_data['analysis_end_time'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    with open(os.path.join(outdir,"results.json"),"w") as oh:
        oh.write(json.dumps(run_data,indent=4))


# call main function
if __name__ == '__main__':
    run()






from locidex.classes import run_command

class mafft:
    input_fasta = ''
    params = {}
    alignment = {
        'params':{},
        'alignment':{}
    }
    status = True
    messages = []

    def __init__(self,input_fasta,params={}):
        (stdout, stderr) = self.align(input_fasta,params)
        self.alignment['params'] = params
        self.alignment['alignment'] = self.parse_align(stdout)
        self.messages.append(stderr)
        if len(self.alignment['alignment']) == 0:
            self.status = False


    def get_alignment(self):
        return self.alignment

    def align(self,input_fasta,params):
        command = ['mafft']
        for p in params:
            command.append(f'--{p}')
            command.append(params[p])
        command.append(input_fasta)
        command = " ".join([str(x) for x in command])
        (stdout, stderr) = run_command(command)
        return (stdout,stderr)


    def parse_align(self,align):
        align = align.split('\n')
        seqs = {}
        for row in align:
            if len(row) == 0:
                continue
            if row[0] == '>':
                id = row[1:]
                seqs[id] = []
                continue
            seqs[id].append(row.lower())
        for id in seqs:
            seqs[id] = "".join([str(x) for x in seqs[id]])
        return seqs









import os
import sys

from locidex.classes.gbk import parse_gbk
from locidex.classes.fasta import ParseFasta
from locidex.utils import guess_alphabet, calc_md5, six_frame_translation, slots
from locidex.classes.prodigal import gene_prediction
from locidex.constants import DNA_AMBIG_CHARS, DNA_IUPAC_CHARS, CharacterConstants, DBConfig, raise_file_not_found_e
import logging
from typing import NamedTuple, Optional, List
from dataclasses import dataclass, asdict

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.INFO)

@dataclass
class HitFilters:
    min_dna_len: int
    max_dna_len: int
    min_dna_ident: float
    min_dna_match_cov: float
    min_aa_len: int
    max_aa_len: int
    min_aa_ident: float
    min_aa_match_cov: float
    dna_ambig_count: int
    __slots__ = slots(__annotations__)

@dataclass
class SeqObject:
    parent_id: str
    locus_name: str
    seq_id: str
    dna_seq: str
    dna_ambig_count: int
    dna_hash: str
    dna_len: int
    aa_seq: str
    aa_hash: str
    aa_len: int
    start_codon: Optional[str] 
    end_codon: Optional[str]
    count_internal_stop: Optional[int]
    # Manually adding slots for compatibility
    __slots__ = slots(__annotations__)

    def to_dict(self) -> dict:
        return asdict(self)

class seq_intake:
    valid_types = ['genbank','gff','gtf','fasta']
    is_file_valid = ''


    def __init__(self,input_file,file_type,feat_key='CDS',translation_table=11,perform_annotation=False,num_threads=1,skip_trans=False):
        if not input_file.exists():
            logger.critical("Seq_intake file {} could not be found".format(input_file))
            raise_file_not_found_e(input_file, logger)
        
        self.input_file = input_file
        self.file_type = file_type
        self.translation_table = translation_table
        self.feat_key = feat_key
        self.skip_trans = skip_trans
        self.num_threads = num_threads
        self.prodigal_genes = []
        self.skip_trans = False
        #self.seq_data = self.process_fasta()
        self.status = True


        if file_type == 'genbank':
            self.seq_data = self.process_gbk()
            self.status = True
        elif file_type == 'fasta' and perform_annotation==True:
            self.seq_data = self.annotate_fasta(self.input_file, num_threads=self.num_threads)
        elif file_type == 'fasta' and perform_annotation==False:
            self.seq_data = self.process_fasta()
        elif file_type == 'gff': # TODO these lists do not contain all allowed file types
            self.status = False
        elif file_type == 'gtf':
            self.status = False
        else:
            raise AttributeError

        if self.status:
            self.add_codon_data()

    
    def annotate_fasta(self, input_file, num_threads):
        sobj = gene_prediction(input_file)
        sobj.predict(num_threads)
        self.prodigal_genes = sobj.genes
        seq_data = self.process_seq_hash(sobj.sequences)
        return self.process_fasta(seq_data)

    def add_codon_data(self):
        for record in self.seq_data:
            if record.aa_len == 0:
                continue
            dna_seq = record.dna_seq.lower().replace('-','')
            dna_len = len(dna_seq)
            start_codon = ''
            stop_codon = ''
            count_internal_stop = 0
            if dna_len >= 6:
                start_codon = dna_seq[0:3]
                stop_codon = dna_seq[-3:]
                count_internal_stop = record.aa_seq[:-1].count(CharacterConstants.stop_codon)
            record.start_codon = start_codon
            record.end_codon = stop_codon
            record.count_internal_stop = count_internal_stop


    def process_gbk(self) -> list[SeqObject]:
        obj = parse_gbk(self.input_file)
        seq_data = []
        if obj.status == False:
            return False
        acs = obj.get_acs()

        for a in acs:
            features = obj.get_feature(a,self.feat_key)
            for seq in features:
                s = seq['dna_seq'].lower().replace('-','')
                for char in DNA_IUPAC_CHARS:
                    s = s.replace(char,"n")

                seq_data.append(SeqObject(
                    parent_id = a,
                    locus_name = seq['gene_name'],
                    seq_id = seq['gene_name'],
                    dna_seq = s,
                    dna_ambig_count = self.count_ambig_chars(seq['dna_seq'], DNA_AMBIG_CHARS),
                    dna_hash = calc_md5([s])[0],
                    dna_len = len(seq['dna_seq']),
                    aa_seq = seq['aa_seq'],
                    aa_hash = seq['aa_hash'],
                    aa_len = len(seq['aa_seq']),
                    start_codon=None,
                    end_codon=None,
                    count_internal_stop=None,
                ))
        return seq_data

    def process_fasta(self, seq_data = []) -> list[SeqObject]:
        obj = ParseFasta(self.input_file)
        ids = obj.get_seqids()
        for id in ids:
            features = obj.get_seq_by_id(id)
            seq = features.seq
            dtype = guess_alphabet(seq)
            dna_seq = ''
            dna_hash = ''
            dna_len = 0
            aa_seq = ''
            aa_hash = ''
            aa_len = 0
            if dtype == 'dna':
                dna_seq = seq
                for char in DNA_IUPAC_CHARS:
                    dna_seq = dna_seq.replace(char,"n")
                dna_hash = calc_md5([seq])[0]
                dna_len = len(seq)
                if self.skip_trans:
                    aa_seq = ''
                else:
                    aa_seq = six_frame_translation(dna_seq,self.translation_table)[0]

                aa_hash = calc_md5([aa_seq])[0]
                aa_len = len(aa_seq)
            else:
                aa_seq = seq
                aa_hash = calc_md5([aa_seq])[0]
                aa_len = len(aa_seq)

            seq_data.append(SeqObject(
                parent_id=features.gene_name,
                locus_name = features.gene_name,
                seq_id = features.seq_id,
                dna_seq = dna_seq,
                dna_ambig_count = self.count_ambig_chars(dna_seq, DNA_AMBIG_CHARS),
                dna_hash = dna_hash,
                dna_len = dna_len,
                aa_seq = aa_seq,
                aa_hash = aa_hash,
                aa_len = aa_len,
                start_codon=None,
                end_codon=None,
                count_internal_stop=None,))
        return seq_data


    def process_seq_hash(self,sequences) -> list[SeqObject]:
        seq_data = []
        for id in sequences:
            seq = sequences[id]
            dtype = guess_alphabet(seq)
            dna_seq = ''
            dna_hash = ''
            dna_len = 0
            aa_seq = ''
            aa_hash = ''
            aa_len = 0
            if dtype == 'dna':
                dna_seq = seq.lower().replace('-','')
                for char in DNA_IUPAC_CHARS:
                    dna_seq = dna_seq.replace(char,"n")
                dna_hash = calc_md5([seq])[0]
                dna_len = len(seq)
                if self.skip_trans:
                    aa_seq = ''
                else:
                    aa_seq = six_frame_translation(dna_seq, self.translation_table)[0]
                aa_hash = calc_md5([aa_seq])[0]
                aa_len = len(aa_seq)
            else:
                aa_seq = seq.lower().replace('-','')
                aa_hash = calc_md5([aa_seq])[0]
                aa_len = len(aa_seq)
            seq_data.append(SeqObject(
                parent_id=id,
                locus_name=id,
                seq_id=id,
                dna_seq= dna_seq,
                dna_hash= dna_hash,
                dna_ambig_count=self.count_ambig_chars(dna_seq, DNA_AMBIG_CHARS),
                dna_len=dna_len,
                aa_seq=aa_seq,
                aa_hash=aa_hash,
                aa_len=aa_len,
                start_codon=None,
                end_codon=None,
                count_internal_stop=None,
            ))

        return seq_data

    def count_ambig_chars(self,seq,chars):
        count = 0
        for char in chars:
            count+= seq.count(char)
        return count


class seq_store:
    #stored_fields = ['parent_id','locus_name','seq_id','dna_hash','dna_len','aa_hash','aa_len','start_codon','stop_codon','count_internal_stop','dna_ambig_count']
    record = {
        'db_info': {},
        'db_seq_info': {},
        'query_data': {
            'sample_name':'',
            'query_seq_data': {},
            'query_hit_columns': [],
            'query_hits': {},
            "locus_profile":{}
        }
    }

    #def __init__(self,sample_name,db_config_dict,metadata_dict,query_seq_records,blast_columns,filters={},stored_fields=[]):
    def __init__(self,sample_name,db_config_dict,metadata_dict,query_seq_records,blast_columns,filters: HitFilters,override=False):
        self.sample_name = sample_name
        self.override = override
        self.record['query_data']['sample_name'] = sample_name
        self.add_db_config(db_config_dict)
        self.add_seq_data(query_seq_records)
        self.add_db_metadata(metadata_dict)
        self.add_hit_cols(blast_columns)
        self.filters = asdict(filters)
        self.record['query_data']['sample_name'] = self.sample_name
       


    def add_db_config(self,conf: DBConfig):
        self.record['db_info'] = conf.to_dict()

    def add_hit_cols(self,columns):
        self.record['query_hit_columns'] = columns

    def add_seq_data(self,query_seq_records: List[SeqObject]):
        for idx, v in enumerate(query_seq_records):
            self.record['query_data']['query_seq_data'][idx] = v.to_dict()

    def add_db_metadata(self,metadata_dict):
        locus_profile = {}
        for seq_index in metadata_dict:
            self.record['db_seq_info'][seq_index] = metadata_dict[seq_index]
            locus_profile[metadata_dict[seq_index]['locus_name']] = {
                    'nucleotide':set(),
                    'protein':set()
                }
        self.record['query_data']["locus_profile"] = locus_profile

    def prepopulate_hit_data(self,df,id_col):
        ids = list(df[id_col].unique())
        for i in ids:
            if str(i) not in self.record['query_data']['query_hits']:
                self.record['query_data']['query_hits'][str(i)] = {
                    'nucleotide':[],
                    'protein':[]
                }

    def add_hit_data(self,df,dbtype,id_col,prepopulate=True):
        if prepopulate:
            self.prepopulate_hit_data(df,id_col)
        for index, row in df.iterrows():
            qid = str(row[id_col])
            self.record['query_data']['query_hits'][qid][dbtype].append(row.to_dict())

    def convert_profile_to_list(self):
        for locus_name in self.record['query_data']['locus_profile']:
            for dtype in self.record['query_data']['locus_profile'][locus_name]:
                self.record['query_data']['locus_profile'][locus_name][dtype] = list(self.record['query_data']['locus_profile'][locus_name][dtype])

    def filter_hits(self):
        query_hits = self.record['query_data']['query_hits']
        for qid in query_hits:
            for dbtype in query_hits[qid]:
                filt = []
                pbitscore = 0
                pbest_hit = ''
                for hit in query_hits[qid][dbtype]:
                    hit_id = str(hit['sseqid'])

                    qlen = hit['qlen']
                    pident = hit['pident']
                    qcovs = hit['qcovs']
                    bitscore = hit['bitscore']
                    hit_name = self.record['db_seq_info'][hit_id]['locus_name']

                    hinfo = self.record["db_seq_info"][hit_id]
                    if self.override:
                        min_len = self.filters["min_dna_len"]
                        min_cov = self.filters["min_dna_match_cov"]
                        min_ident = self.filters["min_dna_ident"]
                        max_len = self.filters["max_dna_len"]
                    else:
                        if dbtype == 'nucleotide':
                            if "dna_min_len" not in hinfo:
                                min_len = self.filters["dna_min_len"]
                            else:
                                min_len = hinfo["dna_min_len"]
                            if "dna_max_len" not in hinfo:
                                max_len = self.filters["dna_min_len"]
                            else:
                                max_len = hinfo["dna_max_len"]
                            if "min_dna_match_cov" not in hinfo:
                                min_cov = self.filters['min_dna_match_cov']
                            else:
                                min_cov = hinfo["min_dna_match_cov"]
                            if "dna_min_ident" not in hinfo:
                                min_ident = self.filters["dna_min_ident"]
                            else:
                                min_ident = hinfo["dna_min_ident"]
                        else:
                            if qlen < hinfo["aa_min_len"] or qlen > hinfo["aa_max_len"] or pident < hinfo["aa_min_ident"]:
                                continue
                            if "aa_min_len" not in hinfo:
                                min_len = self.filters["aa_min_len"]
                            else:
                                min_len = hinfo["aa_min_len"]
                            if "aa_max_len" not in hinfo:
                                max_len = self.filters["aa_min_len"]
                            else:
                                max_len = hinfo["aa_max_len"]
                            if "min_aa_match_cov" not in hinfo:
                                min_cov = self.filters["aa_min_cov"]
                            else:
                                min_cov = hinfo["min_aa_match_cov"]
                            if "aa_min_ident" not in hinfo:
                                min_ident = self.filters["aa_min_ident"]
                            else:
                                min_ident = hinfo["aa_min_ident"]

                    if qlen < min_len or qlen > max_len or pident < min_ident or qcovs < min_cov:
                        continue
                    self.record['query_data']["locus_profile"][hit_name][dbtype].add(qid)
                    filt.append(hit)
                    if bitscore > pbitscore:
                        pbitscore = bitscore
                        pbest_hit = hit
                query_hits[qid][dbtype] = filt









import pytest
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
import gzip
from locidex.classes.gbk import parse_gbk

# Helper to create a GenBank file
def create_genbank_file(path, seq_records, compress=False):
    mode = 'wt'
    _open = gzip.open if compress else open
    with _open(path, mode) as f:
        for record in seq_records:
            # Ensure the record includes necessary annotations for GenBank
            record.annotations["molecule_type"] = "DNA"
        SeqIO.write(seq_records, f, "genbank")

@pytest.fixture
def genbank_file(tmp_path):
    # Create a GenBank file for testing
    file_path = tmp_path / "test.gbk"
    seqs = [SeqRecord(Seq("ATGGAATAG"), id="test_seq1", description="A simple sequence", annotations={"molecule_type": "DNA"})]
    create_genbank_file(file_path, seqs)
    return file_path

@pytest.fixture
def gzip_genbank_file(tmp_path):
    # Create a gzip-compressed GenBank file
    file_path = tmp_path / "test_compressed.gbk.gz"
    seqs = [SeqRecord(Seq("ATGGAATAG"), id="test_seq1", description="A simple sequence", annotations={"molecule_type": "DNA"})]
    create_genbank_file(file_path, seqs, compress=True)
    return file_path

def test_initialization_nonexistent_file():
    # Ensure it handles non-existent files correctly
    with pytest.raises(FileNotFoundError):
        parse_gbk("nonexistent.gbk")

def test_initialization_existing_file(genbank_file):
    # Test initialization with an existing file
    parser = parse_gbk(str(genbank_file))
    assert parser.status, "Parser should be successfully initialized with a valid file"

def test_parse_reference_gbk(genbank_file):
    # Test the parsing functionality
    parser = parse_gbk(str(genbank_file))
    parser.parse_reference_gbk()
    assert len(parser.seq_obj) > 0, "Should parse GenBank file and store data"

def test_get_acs(genbank_file):
    # Test getting accession numbers
    parser = parse_gbk(str(genbank_file))
    acs = parser.get_acs()
    assert len(acs) > 0, "Should return list of accession numbers"

def test_get_seq_by_acs(genbank_file):
    # Test fetching sequences by accession number
    parser = parse_gbk(str(genbank_file))
    seq_data = parser.get_seq_by_acs('test_seq1')
    assert seq_data, "Should return sequence data for given accession"

def test_get_feature(genbank_file):
    # Test fetching features by accession
    parser = parse_gbk(str(genbank_file))
    feature_data = parser.get_feature('test_seq1', 'source')
    assert feature_data, "Should return specific feature data for given accession"

def test_gene_prediction_gzip_fasta(gzip_genbank_file):
    # Ensure it can handle gzip compressed files
    parser = parse_gbk(str(gzip_genbank_file))
    assert parser.status, "Should be able to handle gzip compressed GenBank files"
    assert len(parser.seq_obj) > 0, "Should parse compressed GenBank file and store data"



import pytest
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import tempfile
from pathlib import Path
import os
from locidex.classes.fasta import ParseFasta

@pytest.fixture
def fasta_content():
    return [
        SeqRecord(Seq("ATGCGTACGTAGCTAGC"), id="gene1|123", description=""),
        SeqRecord(Seq("CGTAGCTAGCTGACGTA"), id="gene2|456", description=""),
    ]

@pytest.fixture
def fasta_file(fasta_content):
    # Create a temporary FASTA file in text mode
    with tempfile.NamedTemporaryFile(mode='w+', delete=False, suffix=".fasta") as tmp:
        SeqIO.write(fasta_content, tmp, "fasta")
        tmp_path = tmp.name
    yield Path(tmp_path)
    os.unlink(tmp_path)

def test_parse_fasta_normal(fasta_file):
    parser = ParseFasta(fasta_file)
    assert len(parser.get_seqids()) == 2, "There should be two sequences"
    seq_data = parser.get_seq_by_id("gene1|123")
    assert seq_data.seq == "ATGCGTACGTAGCTAGC".lower(), "Sequence data should match the input"

def test_parse_fasta_with_nonexistent_file():
    # Assuming that the locidex errors out with File not found error if FASTA is non existant:
    with pytest.raises(FileNotFoundError):
        ParseFasta(Path("nonexistent.fasta"))

    # parser = parse_fasta("nonexistent.fasta")
    # assert not parser.status, "Parser status should be False when file does not exist"

def test_parse_fasta_with_definitions(fasta_file):
    parser = ParseFasta(fasta_file, parse_def=True, delim="|")
    seq_data = parser.get_seq_by_id("gene1|123")
    assert seq_data.gene_name == "gene1", "Gene name should be correctly parsed from definition"
    assert seq_data.seq_id == "123", "Seq ID should be correctly parsed from definition"



import os, warnings
import locidex
from locidex.classes.seq_intake import SeqObject, seq_intake, seq_store, HitFilters
from locidex.constants import BlastColumns, DB_EXPECTED_FILES, DBConfig
from locidex.classes.db import search_db_conf, db_config
from collections import Counter
from dataclasses import asdict
from pathlib import Path

PACKAGE_ROOT = os.path.dirname(locidex.__file__)


def seq_intake_class_init(input_file, file_type, perform_annotation):
   #reset global class variables to avoid ambiguous results

   obj = seq_intake(input_file=Path(input_file),
                                      file_type=file_type,feat_key='CDS',translation_table=11,
                                      perform_annotation=perform_annotation,num_threads=1,skip_trans=False)
   return obj

#@pytest.mark.skip(reason="no way of currently testing this")
def test_seq_store_class():
    db_dir = os.path.join(PACKAGE_ROOT, 'example/build_db_mlst_out')
    db_database_config = search_db_conf(db_dir, DB_EXPECTED_FILES, DBConfig._keys())
    metadata_obj = db_config(db_database_config.meta_file_path, ['meta', 'info'])
    sample_name = 'NC_003198.1.fasta'

    seq_obj = seq_intake(input_file=Path(os.path.join(PACKAGE_ROOT, 'example/search/NC_003198.1.fasta')),
                          file_type='fasta', perform_annotation=False)
    
    hit_filters = HitFilters(**{'min_dna_len': 1, 'max_dna_len': 10000000, 'min_dna_ident': 80.0, 'min_dna_match_cov': 80.0, 'min_aa_len': 1, 
                   'max_aa_len': 10000000, 'min_aa_ident': 80.0, 'min_aa_match_cov': 80.0, 'dna_ambig_count': 99999999999999})
    
    seq_store_obj = seq_store(sample_name, DBConfig(**db_database_config.config_obj.config), metadata_obj.config['meta'],
                          seq_obj.seq_data, BlastColumns._fields, hit_filters)
    
    assert list(seq_store_obj.record.keys()) == ['db_info', 'db_seq_info', 'query_data', 'query_hit_columns']
    assert list(seq_store_obj.record['db_info'].keys()) == ['db_name', 'db_version', 'db_date', 'db_author', 
                                                            'db_desc', 'db_num_seqs', 'is_nucl', 'is_prot', 'nucleotide_db_name', 'protein_db_name']
    assert seq_store_obj.record['query_data']['sample_name'] == sample_name
    if len(seq_store_obj.record['query_data']['query_seq_data']) == 1:
        assert len(seq_store_obj.record['query_data']['query_seq_data']) == 1
    else:    
        warnings.warn(f"expected len(seq_store_obj.record['query_data']['query_seq_data']) == 1 but got {len(seq_store_obj.record['query_data']['query_seq_data'])}")
    
    compare_dict = seq_store_obj.record['query_data']['query_seq_data'][0]
    assert set(compare_dict.keys()) ==  set(['parent_id', 'locus_name', 'seq_id', 'dna_hash', 'dna_len', 'aa_hash', 
                                                                                    'aa_len', 'start_codon', 'end_codon', 'count_internal_stop', 'dna_ambig_count', 'dna_seq', 'aa_seq'])
    assert list(seq_store_obj.record['query_data']['locus_profile'].keys()) == ['aroC', 'dnaN', 'hemD', 'hisD', 'purE', 'sucA', 'thrA']
    assert seq_store_obj.record['query_data']['query_hit_columns'] == []
    assert seq_store_obj.record['query_data']['query_hits'] == {}


def test_read_gbk_file():
    seq_intake_object = seq_intake_class_init(input_file=os.path.join(PACKAGE_ROOT, 'example/search/NC_003198.1.gbk'),
                        file_type='genbank', perform_annotation=False)
    
    expected_orfs = 4644
    assert seq_intake_object.file_type == 'genbank'
    assert seq_intake_object.status == True
    assert seq_intake_object.translation_table == 11

    if len(seq_intake_object.seq_data) != expected_orfs:
        msg = f"Expected ORFs number is {expected_orfs} but found {len(seq_intake_object.seq_data)}! Check class init"
        warnings.warn(msg)
    else:
        assert len(seq_intake_object.seq_data) == expected_orfs

    assert seq_intake_object.seq_data[0] == SeqObject(**{'parent_id': 'NC_003198', 'locus_name': 'STY_RS00005', 'seq_id': 'STY_RS00005', 
            'dna_seq': 'atgaaccgcatcagcaccaccaccattaccaccatcaccattaccacaggtaacggtgcgggctga', 
            'dna_hash': 'f46b7aac05dba47f42391aaa5ac25edf', 'count_internal_stop': 0, 'dna_ambig_count': 0, 'start_codon': 'atg', 'end_codon': 'tga',
            'dna_len': 66, 'aa_seq': 'mnristttittitittgngag', 'aa_hash': '8b370db9e32fd0a8362c35f3535303d8', 'aa_len': 21})
    assert all([record.start_codon for record in seq_intake_object.seq_data if record.aa_len > 0]) == True
    #assert dict(Counter(['start_codon' in record if record.aa_len > 0 else False in record for record in seq_intake_object.seq_data ])) == {True: 4325, False: 319}
    assert dict(Counter([record.start_codon is not None if record.aa_len > 0 else False for record in seq_intake_object.seq_data ])) == {True: 4325, False: 319}
    

def test_read_fasta_file():
    expected_orfs=4653
    seq_intake_object = seq_intake_class_init(input_file=Path(os.path.join(PACKAGE_ROOT, 'example/search/NC_003198.1.fasta')),
                          file_type='fasta', perform_annotation=True)
    assert seq_intake_object.file_type == 'fasta'
    assert seq_intake_object.feat_key == 'CDS'
    assert seq_intake_object.status == True
    assert seq_intake_object.translation_table == 11
    assert seq_intake_object.valid_types == ['genbank', 'gff', 'gtf', 'fasta']
    if len(seq_intake_object.seq_data) != expected_orfs:
        msg = f"Expected ORFs number is {expected_orfs} but found {len(seq_intake_object.seq_data)}! Check pyrodigal and python versions."
        warnings.warn(msg)
    assert len(seq_intake_object.seq_data) > 0
    assert sum([contig.aa_len for contig in seq_intake_object.seq_data]) > 0
    assert any([True if 'NC_003198.1' in contig.parent_id else False  for contig in seq_intake_object.seq_data]) == True
    assert any([True if 'NC_003198.1' in contig.locus_name else False  for contig in seq_intake_object.seq_data]) == True
    assert any([True if 'NC_003198.1' in contig.seq_id else False  for contig in seq_intake_object.seq_data]) == True
    assert len(seq_intake_object.seq_data[0].dna_seq) > 0
    assert seq_intake_object.seq_data[0].dna_len > 0
    assert len(seq_intake_object.seq_data[0].dna_hash) > 0
    assert any([ True if contig.dna_ambig_count == 0 else False  for contig in seq_intake_object.seq_data]) == True






import pathlib
import json
from typing import List, Union, Tuple, Dict, Optional
from dataclasses import dataclass
import os
import re
import sys
from argparse import (ArgumentParser, ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter)
from datetime import datetime
from locidex.version import __version__
from locidex.constants import DBConfig, DBFiles, raise_file_not_found_e
import logging
import errno

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.DEBUG)

class DBData:
    """
    Validate and get all database data for other modules.
    
    * This class will create some redundancy and will need to be refactored to reflect
    * the overall use of this module in the future. But once refactoring is complete 
    * it should be much easier to refactor this and the other modules. Additionally, at that
    * point we should have a better understanding of how all the modules fit together.
    """

    __nucleotide_name = "nucleotide"
    __nucleotide_db_type = "nucl"
    __protein_name = "protein"
    __protein_db_type = "prot"
    __nucleotide_path = pathlib.Path(__nucleotide_name)
    __protein_path = pathlib.Path(__protein_name)

    def __init__(self, db_dir: pathlib.Path):
        self.db_dir = pathlib.Path(db_dir)
        self.config_data: DBConfig = self._get_config(self.db_dir)
        self.metadata: dict = self._get_metadata(self.db_dir)
        self.nucleotide, self.protein = self._get_blast_dbs(self.db_dir, self.config_data)

    @classmethod
    def nucleotide_db_type(cls):
        return cls.__nucleotide_db_type

    @classmethod
    def protein_db_type(cls):
        return cls.__protein_db_type

    @classmethod
    def protein_name(cls):
        return cls.__protein_name
    
    @classmethod
    def nucleotide_name(cls):
        return cls.__nucleotide_name

    @property
    def nucleotide_blast_db(self):
        if self.nucleotide is None:
            logger.critical("Nucleotide blast database does not exist.")
            raise_file_not_found_e(self.nucleotide, logger)
        return self.nucleotide / self.__nucleotide_path
    
    @property
    def protein_blast_db(self):
        if self.protein is None:
            logger.critical("Protein blast database does not exist.")
            raise_file_not_found_e(self.protein, logger)
        return self.protein / self.__protein_path

    def _get_config(self, db_dir: pathlib.Path) -> DBConfig:
        """
        Validates the config file and searializes the data into a DBConfig object
        """
        return check_config(db_dir)
    
    def _get_metadata(self, db_dir: pathlib.Path) -> dict:
        metadata_file = db_dir.joinpath(DBFiles.meta_file)
        if not metadata_file.exists():
            logger.critical("Metadata file does not appear to exist in db: {}".format(db_dir))
            raise_file_not_found_e(str(metadata_file), logger)
        md_data = None
        with open(metadata_file, 'r') as md:
            md_data = json.load(md)
        return md_data


    def _get_blast_dbs(self, db_dir: pathlib.Path, config_data: DBConfig) -> Tuple[Optional[pathlib.Path], Optional[pathlib.Path]]:
        blast_db = db_dir.joinpath(DBFiles.blast_dir)
        nucleotide: Optional[pathlib.Path] = None
        protein: Optional[pathlib.Path] = None
        if not blast_db.exists():
            logger.critical("blast directory not found. Database path maybe incorrect: {}".format(str(db_dir)))
            raise NotADirectoryError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), str(db_dir))
        if config_data.is_nucl:
            nucleotide = blast_db.joinpath(self.__nucleotide_path)
            if not nucleotide.exists():
                logger.critical("Cannot find nucleotide database, but it should exist. {}".format(nucleotide))
                raise_file_not_found_e(nucleotide, logger)
        if config_data.is_prot:
            protein = blast_db.joinpath(self.__protein_path)
            if not protein.exists():
                logger.critical("Cannot find protein database, but it should exist. {}".format(protein))
                raise_file_not_found_e(protein, logger)
        return nucleotide, protein


class ManifestItem:
    """
    Manifest item created for exporting and importing locidex items
    """
    __path_key = 'path'
    __config_key = 'config'

    def __init__(self, db: pathlib.Path, root_db: pathlib.Path, config: DBConfig):
        """
        db Path: Relative path to the allele database
        root_db Path: The directory containing the manifest.json file required to resolve paths
        config DBConfig: Database configuration data
        """
        self.db = db
        self.config = config
        self.root_db = root_db

    @property
    def db_path(self):
        return self.root_db / self.db

    def to_dict(self):
        return {self.__path_key: str(self.db), self.__config_key: self.config.to_dict()}
    
    def __repr__(self) -> str:
        return "Allele location: {}\n Config data: {}".format(self.db, self.config)
    
    @classmethod
    def path_key(cls):
        """
        ! Not passing this as a property for the class method as apparently that will be deprecated in 3.13
        """
        return cls.__path_key
    
    @classmethod
    def config_key(cls):
        """
        ! Not passing this as a property for the class method as apparently that will be deprecated in 3.13
        """
        return cls.__config_key


@dataclass(frozen=True)
class _Constants:
    manifest_name: pathlib.Path = "manifest.json"

def add_args(parser=None):
    if parser is None:
        parser = ArgumentParser(
            description="Locidex manifest: Setup directory of databases for use with search")
    parser.add_argument('-i','--input', type=str, required=True,help='Input directory containing multiple locidex databases')
    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)
    return parser


def check_config(directory: pathlib.Path) -> DBConfig:
    """
    Validate config file in a directory. Throws an error if any required parameters
    are missing.

    directory: Path of the directory containing the parent.
    """

    config_dir = pathlib.Path(directory).joinpath(DBFiles.config_file)
    config_data: Optional[DBConfig] = None
    if not config_dir.exists():
        logger.critical("Could not find config file: {}".format(config_dir))
        raise_file_not_found_e(config_dir, logger)
    
    with open(config_dir, 'r') as conf:
        config_data = DBConfig(**json.load(conf))
        for k, v in config_data.to_dict().items():
            if v is None or v == '':
                logger.critical("Missing value in config file for key {} which has value {}".format(k, v))
                raise AttributeError("Config cannot have missing values: {}".format(k))
    return config_data

def validate_db_files(allele_dir: List[pathlib.Path], file_in: pathlib.Path) -> List[Tuple[pathlib.Path, DBConfig]]:
    """
    Validates a directory of allele databases, and verifies that the config contains
    the required fields

    allele_dir List[pathlib.Path: Directory of various allele databases needed by locidex
    file_in Path: Root directory to set files relative too
    """
    db_configs: Tuple[pathlib.Path, DBConfig] = []
    for a_dir in allele_dir:
        for k, v in DBFiles.items():
            if not pathlib.Path(a_dir / v).exists():
                logger.critical("Required file {} does not exist.".format(k))
                raise_file_not_found_e(str(a_dir / v), logger)
            
        db_configs.append((a_dir.relative_to(file_in), check_config(a_dir)))
    return db_configs


def check_dbs(file_in: pathlib.Path) -> List[pathlib.PosixPath]:
    """
    Checks that all locidex databases in a directory are complete.

    file_in: A path to a directory of databases
    """
    logger.debug("Checking that the following databases exist in: {}".format(file_in))
    db_dirs = [p for p in file_in.iterdir() if p.is_dir()]
    if not db_dirs:
        logger.critical("No valid databases found in: {}".format(file_in))
        raise AssertionError("No valid databases found in: {}".format(file_in))
    return db_dirs

def create_manifest(file_in: pathlib.Path) -> Dict[str, List[Dict[str, str]]]:
    """
    Create a manifest file for each of the locidex dbs.

    file_in pathlib.Path: File path to directory of databases
    """
    allele_dirs: List[pathlib.Path] = check_dbs(file_in)
    validated_dbs: List[Tuple[pathlib.Path, DBConfig]] = validate_db_files(allele_dirs, file_in)
    db_manifest = dict()
    for path, conf in validated_dbs:
        logger.info("Adding database: {} to manifest.".format(str(path)))
        if db_manifest.get(conf.db_name) is None:
            db_manifest[conf.db_name] = []
            
        if db_manifest[conf.db_name] and (versions := [i.db_version for i in db_manifest[conf.db_name]]):
            if conf.db_version in versions:
                logger.critical("Databases with the same name and version have been specified (name: {}, path: {}, version: {})".format(conf.db_name, path, conf.db_version))
                raise KeyError("Databases with the same name and version have been specified (name: {}, path: {}, version: {})".format(conf.db_name, path, conf.db_version))
            
        db_manifest[conf.db_name].append(ManifestItem(db=path, config=conf, root_db=file_in).to_dict())
    return db_manifest


def write_manifest(file_in: pathlib.Path, manifest: Dict[str, List[Dict[str, str]]]) -> pathlib.Path:
    """
    Write the manifest.json file

    file_in Path: Specified input directory
    manifest dict: data to write to the manifest
    """

    manifest_file = _Constants.manifest_name
    logger.debug("Creating manifest file in directory: {} with name: {}".format(file_in, manifest_file))
    path_out = file_in.joinpath(manifest_file)
    logger.info("Writing manifest file to: {}".format(str(path_out)))
    with open(path_out, 'w', encoding='utf8') as m_out:
        json.dump(manifest, m_out, indent=2)
    return path_out


def run(cmd_args=None):
    if cmd_args is None:
        parser = add_args()
        cmd_args = parser.parse_args()
    directory_in = pathlib.Path(cmd_args.input)
    if not directory_in.exists():
        logger.critical("Directory: {} does not appear to exist.".format(str(directory_in)))
        raise NotADirectoryError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), str(directory_in))

    manifest = create_manifest(directory_in)
    return write_manifest(directory_in, manifest)

def select_db(manifest_data: Dict[str, List[ManifestItem]], name: str, version: str) -> ManifestItem:
    """
    Select a locidex database from the manifest file provided.

    manifest_data Dict[str, List[ManifestItem]]: Parsed manifest file data for selecting a database
    name str: Name of database to select
    version str: version of selected database to select
    """
    db_data = manifest_data.get(name)
    if db_data is None:
        logger.critical("Could not find database with specified name: {}".format(name))
        raise KeyError("Could not find database with specified name: {}".format(name))
    
    try:
        db = next(filter(lambda x: x.config.db_version == version, db_data))
    except StopIteration:
        logger.critical("No database entry with version: {}".format(version))
        raise ValueError("No database entry with version: {}".format(version))
    
    return db

def read_manifest(input_file: pathlib.Path) -> dict:
    """
    input_file Path: Manifest file to be parsed
    """
    if not input_file.is_dir():
        logger.critical("Please pass the database directory, not a file.")
        raise AssertionError("Allele database directory must be passed directly.")
    
    manifest_file = input_file / _Constants.manifest_name
    manifest_data: Dict[str, ManifestItem] = dict()
    with open(manifest_file, 'r', encoding='utf8') as mani_in:
        manifest = json.load(mani_in)
        for k, list_manifests in manifest.items():
            if manifest_data.get(k) is None:
                manifest_data[k] = []
            for v in list_manifests:
                manifest_item = ManifestItem(db=v[ManifestItem.path_key()], config=DBConfig(**v[ManifestItem.config_key()]), root_db=input_file)
                manifest_data[k].append(manifest_item)
    return manifest_data

def get_manifest_db(input_file: pathlib.Path, name: str, version: str) -> ManifestItem:
    """
    Returns path to the database file selected
    """
    output = read_manifest(input_file)
    db_out = select_db(output, name, version)
    return db_out

# call main function
if __name__ == '__main__':
    run()




from Bio import GenBank
import gzip
from mimetypes import guess_type
from functools import partial
import os
from locidex.utils import revcomp,calc_md5

class parse_gbk:
    messages = []

    def __init__(self,input_file):
        self.input_file= input_file
        self.status = True
        if not os.path.isfile(self.input_file):
            self.messages.append(f'Error {self.input_file} does not exist')
            self.status = False
        self.seq_obj = self.parse_reference_gbk()


    def get_acs(self):
        if self.seq_obj is not None:
            return list(self.seq_obj.keys())
        else:
            return []

    def get_seq_by_acs(self,acs):
        if self.seq_obj is not None and acs in self.seq_obj:
            return self.seq_obj[acs]
        else:
            return {}

    def get_feature(self,acs,feature):
        s = self.get_seq_by_acs(acs)
        if feature in s['features']:
            return s['features'][feature]
        else:
            return {}


    def parse_reference_gbk(self):
        """
        Method to parse the GenBank reference file, clean the strings, and
        return the reference features of interest.

        Parameters
        ----------
        gbk_file : str
            The file path to the reference genbank format file with sequence annotations
        Returns
        -------
        dict
            A dictionary of all the reference features
        """
        encoding = guess_type(self.input_file)[1]
        _open = partial(gzip.open, mode='rt') if encoding == 'gzip' else open
        sequences = {}
        with _open(self.input_file) as handle:
            for record in GenBank.parse(handle):
                gb_accession = record.locus
                gb_accession_version = 1
                if len(record.accession) == 2:
                    gb_accession = record.accession[0]
                    gb_accession_version = gb_accession[1]
                # clean the sequence
                genome_seq = repr(record.sequence).replace("\'", '')
                sequences[gb_accession] = {
                    'accession': gb_accession,
                    'version': gb_accession_version,
                    'features': {'source': genome_seq}
                }
                features = record.features
                # retrieve the features if present
                for feat in features:
                    if feat.key == 'CDS' or feat.key == '5\'UTR' or feat.key == '3\'UTR':
                        if not feat.key in sequences[gb_accession]['features']:
                            sequences[gb_accession]['features'][feat.key] = []
                        qualifier = feat.qualifiers
                        positions = []
                        gene_name = ''
                        locus_tag = ''
                        aa = ''
                        # more string cleaning
                        for name in qualifier:
                            if name.key == '/gene=':
                                gene_name = name.value.replace("\"", '').strip()
                            if name.key == '/translation=':
                                aa = name.value.replace("\"", '').strip()
                            if name.key == '/locus_tag=':
                                gene_name = name.value.replace("\"", '').strip()
                                locus_tag = gene_name
                        if locus_tag != '':
                            gene_name = locus_tag
                        locations = feat.location.strip().replace("join(", '').replace(')', '').split(',')
                        seq = []
                        # retreive the locations of the features
                        for location in locations:
                            # more string cleaning
                            location = location.replace('<', '').replace('>', '')
                            if not 'complement' in location:
                                location = location.split('.')
                                start = int(location[0]) - 1
                                end = int(location[2])
                                seq.append(genome_seq[start:end].replace("\'", ''))
                                positions.append([start, end])
                            else:
                                location = location.replace('complement(', '').replace(')', '').split('.')
                                start = int(location[0]) - 1
                                end = int(location[2])
                                seq.append(revcomp(genome_seq[start:end].replace("\'", '')))
                                positions.append([start, end])

                        seq = ''.join(seq)
                        sequences[gb_accession]['features'][feat.key].append(
                            {'gene_name': gene_name, 'dna_seq': seq.lower(), 'aa_seq': aa.lower(), 'positions': positions,
                             'gene_len': len(seq),'aa_hash':calc_md5([aa])[0],'dna_hash':calc_md5([seq])[0]})

        return sequences


import hashlib
import json
import os
import shutil
import argparse
import logging
import errno
import getpass
from collections import Counter
from pathlib import Path
from locidex.manifest import ManifestItem
from Bio.Seq import Seq
from typing import Dict, FrozenSet, Optional, List
from locidex.constants import NT_SUB, PROTEIN_ALPHA, DNA_ALPHA, OPTION_GROUPS, FILE_TYPES, raise_file_not_found_e
import locidex.manifest as manifest 


def get_user() -> str:
    """
    Return the user login or a default value if it cannot be determined
    """
    output: str = "No Author"
    try:
        output = getpass.getuser()
    except KeyError:
        pass
    return output 

def slots(annotations: Dict[str, object]) -> FrozenSet[str]:
    """
    Thank you for this: https://stackoverflow.com/a/63658478
    """
    return frozenset(annotations.keys())

def check_db_groups(analysis_params: dict, cmd_args: argparse.Namespace, param_db: str = "db") -> dict:
    """
    Verify that a locidex database, or database group passed has all of the require parameters
    """
    for opt in OPTION_GROUPS:
        if analysis_params[opt] is not None:
            for option in OPTION_GROUPS[opt]:
                if analysis_params[option] is None:
                    raise AttributeError("Missing required parameter: {}".format(option))
    
    if cmd_args.db_group is not None:
        manifest_data = manifest.get_manifest_db(input_file=Path(cmd_args.db_group), name=cmd_args.db_name, version=cmd_args.db_version)
        analysis_params[param_db] = str(manifest_data.db_path)

    return analysis_params

def revcomp(s):
    """
    Helper method to create the reverse complement nucleotide sequence
    for the sequence passed.

    Parameters
    ----------
    s : str
        The nucleotide sequence to parse

    Returns
    -------
    str
        The reverse complement of the passed nucleotide sequence
    """
    return s.translate(NT_SUB)[::-1]

def calc_md5(list):
    """
    Method to encode the MD5 hash for the input string.

    Parameters
    ----------
    string : srt
    The string to compute the MD5 hash

    Returns
    -------
    hash
        The md5 hash generated
    """
    out = []
    for string in list:
        seq = str(string).encode()
        md5 = hashlib.md5()
        md5.update(seq)
        out.append(md5.hexdigest())
    return out

def translate_dna(dna_seq,trans_table=11):
    l = len(dna_seq)
    r = l % 3
    if r == 0:
        s = dna_seq
    else:
        s = ''.join(list(dna_seq)[:-r])
    return str(Seq(s).translate(table=trans_table))

def six_frame_translation(dna_seq,trans_table) -> List[str]:
    fwd = []
    rev = []
    for i in range(0,3):
        fwd.append(translate_dna(dna_seq[i:], trans_table))
        rev.append(translate_dna(revcomp(dna_seq)[i:], trans_table))
    fwd.extend(rev)
    #return (fwd, rev)
    return fwd

def guess_alphabet(seq):
    seq = seq.lower().replace('-','')
    l = len(seq)
    if l == 0:
        return None
    pr_inf_chars = set(PROTEIN_ALPHA) - set(DNA_ALPHA)
    counts = dict(Counter(seq))
    dtype = None
    if len(set(counts.keys()) & pr_inf_chars) > 0:
        dtype = 'protein'
    else:
        dna_bases = ['a','t','c','g','n']
        l = len(seq)
        c = 0
        for b in counts:
            if b not in dna_bases:
                continue
            c+= counts[b]
        if c / l >= 0.7:
            dtype = 'dna'
        else:
            dtype = 'protein'
    return dtype

def write_seq_list(seqs,output_file,format='json',seq_type='dna',seq_id_key='index'):
    with open(output_file, 'w') as oh:
        if format == 'json':
            json.dump(seqs, oh, indent=2)
        elif format == 'fasta':
            i = 0
            for record in seqs:
                if seq_id_key == 'index':
                    id = i
                else:
                    id = record[seq_id_key]
                s = record['dna_seq']
                if seq_type == 'aa':
                    s = record['aa_seq']
                oh.write(f">{id}\n{s}\n")

def write_seq_dict(data,output_file):
    with open(output_file, 'w') as oh:
        for fid, seq in data.items():
            oh.write(f">{fid}\n{seq}\n")
    return output_file


def validate_dict_keys(data_dict,required_keys):
    for k in required_keys:
        if not k in data_dict:
            return [False, k]
    return [True, k ]

def get_file_length(self):
    return int(os.popen(f'wc -l {self.file_path}').read().split()[0])

def filter_hsps_df(df):
    return


def get_format(file: Path) -> Optional[str]:
    """
    Return file type based on suffix used
    """
    format: str = None
    file_exts = file.suffixes
    for k, extensions in FILE_TYPES.items():
        for ext in file_exts:
            if ext in extensions:
                format = k
                break
    return format

def check_utility_installed(utility: str) -> Optional[str]:
    """
    utility str: The name of an executable to verify installation of
    return None if the utility is installed | Message if utility is not installed
    """
    out_string: Optional[str]  = None
    utility_path: Optional[str] = shutil.which(utility)
    if utility_path is None:
        out_string = f"Utility {utility} is not installed or is not executable."
    return out_string


def check_utilities(logger: logging.Logger, utilities: List[str]):
    """
    Check all utilites are installed
    """
    missing_utilities: List[str] = []
    for i in utilities:
        output = check_utility_installed(i)
        if output is not None:
            missing_utilities.append(i)
            logger.critical("Missing: {}".format(output))
    
    if missing_utilities:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), ",".join(missing_utilities))



import pandas as pd
import numpy as np
from locidex.constants import NT_SUB, STOP_CODONS, START_CODONS
class extractor:
    
    def __init__(self,df,seq_data,sseqid_col,queryid_col,qstart_col,qend_col,qlen_col,sstart_col,send_col,slen_col,sstrand_col,bitscore_col,overlap_thresh=100,extend_threshold_ratio = 0.2,filter_contig_breaks=False):
        self.seqs = {}
        self.df = df
        self.filter_contig_breaks = filter_contig_breaks
        self.df = self.set_extraction_pos(df, sstart_col, send_col)
        self.is_complete(self.df,qstart_col,qend_col,qlen_col)
        self.is_contig_boundary(self.df,'qstart','qend','ext_start','ext_end',slen_col,'qlen')
        if filter_contig_breaks:
            self.df = df[ (df['is_5prime_boundary'] == False) & (df['is_3prime_boundary'] == False)]
            self.df = self.df.reset_index(drop=True)
        self.set_revcomp(self.df,sstart_col,send_col,sstrand_col)
        pcols = [qstart_col,qend_col,sstart_col,send_col]
        for c in pcols:
            self.df[c] = self.df[c].apply(lambda x: x - 1)
        self.df = self.get_best_hit_query(self.df)
        sort_cols = [sseqid_col, 'locus_name', sstart_col, bitscore_col, send_col]
        ascending_cols = [True, True, True, False, False]
        self.df = self.df.sort_values(sort_cols,ascending=ascending_cols).reset_index(drop=True)
        self.df = self.recursive_filter_redundant_queries(self.df, 'locus_name', sseqid_col, bitscore_col, 
                                                          sort_cols, ascending_cols, overlap_threshold=1)
        self.df = self.extend(self.df,sseqid_col, queryid_col, qstart_col, qend_col, sstart_col,send_col,slen_col, qlen_col, bitscore_col, overlap_threshold=overlap_thresh)  
        self.df = self.set_extraction_pos(self.df, sstart_col, send_col) 
        loci_ranges = self.group_by_locus(self.df,sseqid_col, queryid_col,qlen_col,extend_threshold_ratio)
        self.seqs = self.extract_seq(loci_ranges, seq_data)
        pass

    def is_5prime_complete(self,df,qstart_col):
        df['is_5prime_complete'] = np.where(df[qstart_col] == 1, True, False)

    def is_3prime_complete(self,df,qend_col,qlen_col):
        df['is_3prime_complete'] = np.where(df[qend_col] == df[qlen_col] , True, False)

    def is_complete(self,df,qstart_col,qend_col,qlen_col):
        self.is_5prime_complete(df,qstart_col)
        self.is_3prime_complete(df,qend_col,qlen_col)
        df['is_complete'] = np.where(((df['is_5prime_complete'] == True) &  (df['is_3prime_complete'] == True)), True, False)

    def is_contig_boundary(self,df,qstart_col, qend_col, sstart_col,send_col,slen_col,qlen_col):
        df['is_5prime_boundary'] = np.where((df[sstart_col] == 1) & (df[qstart_col] > 1), True, False)
        df['is_3prime_boundary'] = np.where((df[send_col] == df[slen_col]) & (df[qend_col] < df[qlen_col]), True, False)
        df['is_on_boundary'] = np.where(((df['is_5prime_boundary'] == True) & (df['is_3prime_boundary'] == True)), True, False)

    def set_revcomp(self,df,sstart_col,send_col,strand_col):
        df['reverse'] = np.where(df[sstart_col] > df[send_col], True, False)
        df['complement'] = np.where(df[strand_col] == 'plus', True, False)

    def set_extraction_pos(self,df,start_col,end_col):
        df['ext_start'] = df[start_col]
        df['ext_end'] = df[end_col]
        for idx, row in df.iterrows():
            start = row['ext_start']
            end = row['ext_end']
            if start > end:
                tmp = end
                end = start
                start = tmp
                df.loc[idx, 'ext_start'] = start
                df.loc[idx, 'ext_end'] = end
        return df

    def get_best_hit_query(self,df):
        sort_cols = ['sseqid', 'locus_name', 'bitscore','qlen','qcovhsp','ext_start','ext_end']
        ascending_cols = [True, True, False, False, False,True, False]
        df = df.sort_values(sort_cols,
                                ascending=ascending_cols).reset_index(drop=True)
        loci = list(df['locus_name'].unique())
        filter_df = []
        for locus_name in loci:
            subset = df[df['locus_name'] == locus_name]
            if len(subset) == 1:
                filter_df.append(subset)
                continue
            
            prev_index  = 0
            prev_contig_id = ''
            prev_contig_start = -1
            prev_contig_end = -1
            prev_score = 0
            filter_rows = []
            for idx, row in subset.iterrows():
                contig_id = row['sseqid']
                contig_start = row['ext_start']
                contig_end = row['ext_end']
                score = float(row['bitscore'])

                if contig_id == prev_contig_id:
                    if (contig_start >= prev_contig_start and contig_start <= prev_contig_end) or (
                            contig_end >= prev_contig_start and contig_end <= prev_contig_end):
                        overlap = abs(contig_start - prev_contig_end)

                        if overlap > 1:
                            if prev_score < score:
                                filter_rows.append(prev_index)
                            else:
                                filter_rows.append(idx)
                                continue

                
                prev_index = idx
                prev_contig_id = contig_id
                prev_contig_start = contig_start
                prev_contig_end = contig_end
                prev_score = score
            valid_ids = list( set(subset.index) - set(filter_rows)  )
            filter_df.append(subset.filter(valid_ids, axis=0))

        return pd.concat(filter_df, ignore_index=True)


    def recursive_filter_redundant_queries(self,df, locus_col, sseqid_col, bitscore_col, sort_cols, ascending_cols, overlap_threshold=1):
        size = len(df)
        prev_size = 0
        while size != prev_size:
            df = df.sort_values(sort_cols,
                                ascending=ascending_cols).reset_index(drop=True)
            df = self.remove_redundant_queries(df, locus_col, sseqid_col, bitscore_col, overlap_threshold=overlap_threshold)
            prev_size = size
            size = len(df)

        return df.sort_values(sort_cols,ascending=ascending_cols).reset_index(drop=True)

    def remove_redundant_queries(self,df,locus_col,sseqid_col, bitscore_col, overlap_threshold=1):
        seq_id_list = list(df[sseqid_col].unique())
        filter_df = []
        for seqid in seq_id_list:
            subset = df[df[sseqid_col] == seqid]
            prev_locus_id = ''
            prev_contig_id = ''
            prev_index = -1
            prev_contig_start = -1
            prev_contig_end = -1
            prev_score = 0
            filter_rows = []
            for idx, row in subset.iterrows():
                contig_id = row[sseqid_col]
                contig_start = row['ext_start']
                contig_end = row['ext_end']
                score = float(row[bitscore_col])
                locus_id = row[locus_col]

                if prev_contig_id == '':
                    prev_index = idx
                    prev_contig_id = contig_id
                    prev_contig_start = contig_start
                    prev_contig_end = contig_end
                    prev_score = score
                    prev_locus_id = locus_id
                    continue

                if locus_id == prev_locus_id:
                    if (contig_start >= prev_contig_start and contig_start <= prev_contig_end) or (
                            contig_end >= prev_contig_start and contig_end <= prev_contig_end):
                        overlap = abs(contig_start - prev_contig_end)

                        if overlap > overlap_threshold:
                            if prev_score < score:
                                filter_rows.append(prev_index)
                            else:
                                filter_rows.append(idx)

                prev_index = idx
                prev_contig_id = contig_id
                prev_contig_start = contig_start
                prev_contig_end = contig_end
                prev_score = score
                prev_locus_id = locus_id


            valid_ids = list( set(subset.index) - set(filter_rows)  )

            filter_df.append(subset.filter(valid_ids, axis=0))


        return pd.concat(filter_df, ignore_index=True)

    def fix_postioning(self,df):
        for idx, row in df.iterrows():
            start = row['ext_start']
            end = row['ext_end']
            rev = row['reverse']
            complement = row['complement']

            if rev and complement:
                start += 1
                end += 1
            elif rev:
                start+=1
            elif complement:
                end += 1
            else:
                end += 1

            if row['reverse'] == True:
                tmp = row['ext_end']
                end = row['ext_start']
                start = tmp
            df.loc[idx, 'ext_start'] = start
            df.loc[idx, 'ext_end'] = end

        return df

    def remove_redundant_hits(self,df,seqid_col, bitscore_col, overlap_threshold=1):
        seq_id_list = list(df[seqid_col].unique())
        filter_df = []
        for seqid in seq_id_list:
            subset = df[df[seqid_col] == seqid]
            prev_contig_id = ''
            prev_index = -1
            prev_contig_start = -1
            prev_contig_end = -1
            prev_score = 0
            filter_rows = []
            for idx, row in subset.iterrows():
                contig_id = row[seqid_col]
                contig_start = row['ext_start']
                contig_end = row['ext_end']
                score = float(row[bitscore_col])

                if prev_contig_id == '':
                    prev_index = idx
                    prev_contig_id = contig_id
                    prev_contig_start = contig_start
                    prev_contig_end = contig_end
                    prev_score = score
                    continue

                if (contig_start >= prev_contig_start and contig_start <= prev_contig_end) or (
                        contig_end >= prev_contig_start and contig_end <= prev_contig_end):
                    overlap = abs(contig_start - prev_contig_end)

                    if overlap > overlap_threshold:
                        if prev_score < score:
                            filter_rows.append(prev_index)
                        else:
                            filter_rows.append(idx)

                prev_index = idx
                prev_contig_id = contig_id
                prev_contig_start = contig_start
                prev_contig_end = contig_end
                prev_score = score


            valid_ids = list( set(subset.index) - set(filter_rows)  )

            filter_df.append(subset.filter(valid_ids, axis=0))


        return pd.concat(filter_df, ignore_index=True)

    def remove_redundant_hits_dual_key(self,df,locus_col, seqid_col, bitscore_col, overlap_threshold=1):
        seq_id_list = list(df[seqid_col].unique())
        filter_df = []
        for seqid in seq_id_list:
            subset = df[df[seqid_col] == seqid]
            prev_contig_id = ''
            prev_locus_id = ''
            prev_index = -1
            prev_contig_start = -1
            prev_contig_end = -1
            prev_score = 0
            filter_rows = []
            for idx, row in subset.iterrows():
                contig_id = row[seqid_col]
                contig_start = row['ext_start']
                contig_end = row['ext_end']
                score = float(row[bitscore_col])
                locus_id = row[locus_col]

                if prev_contig_id == '':
                    prev_index = idx
                    prev_contig_id = contig_id
                    prev_contig_start = contig_start
                    prev_contig_end = contig_end
                    prev_score = score
                    prev_locus_id = locus_id
                    continue

                if locus_id == prev_locus_id and contig_id == prev_contig_id: 
                    if (contig_start >= prev_contig_start and contig_start <= prev_contig_end) or (
                            contig_end >= prev_contig_start and contig_end <= prev_contig_end):
                        
                        overlap = abs(contig_start - prev_contig_end)

                        if overlap > overlap_threshold:
                            if prev_score < score:
                                filter_rows.append(prev_index)
                            else:
                                filter_rows.append(idx)

                prev_index = idx
                prev_contig_id = contig_id
                prev_contig_start = contig_start
                prev_contig_end = contig_end
                prev_score = score
                prev_locus_id = locus_id


            valid_ids = list( set(subset.index) - set(filter_rows)  )

            filter_df.append(subset.filter(valid_ids, axis=0))


        return pd.concat(filter_df, ignore_index=True)

    def recursive_filter_overlap_records(self,df, seqid_col, bitscore_col, sort_cols, ascending_cols, overlap_threshold=1):
        size = len(df)
        prev_size = 0
        while size != prev_size:
            df = df.sort_values(sort_cols,
                                ascending=ascending_cols).reset_index(drop=True)
            df = self.remove_redundant_hits(df, seqid_col, bitscore_col, overlap_threshold=overlap_threshold)
            prev_size = size
            size = len(df)
        return df.sort_values(sort_cols,ascending=ascending_cols).reset_index(drop=True)
    
    def recursive_filter_overlap_records_dual_key(self,df, locus_col, seqid_col, bitscore_col, sort_cols, ascending_cols, overlap_threshold=1):
        size = len(df)
        prev_size = 0
        while size != prev_size:
            df = df.sort_values(sort_cols,
                                ascending=ascending_cols).reset_index(drop=True)
            df = self.remove_redundant_hits_dual_key(df, locus_col, seqid_col, bitscore_col, overlap_threshold=overlap_threshold)
            prev_size = size
            size = len(df)
        return df.sort_values(sort_cols,ascending=ascending_cols).reset_index(drop=True)

    def extract_seq(self,loci_data,seq_data):
        seqs = []
        id = 0
        for locus_name in loci_data:
            for row in loci_data[locus_name]:
                query_id = str(row['query_id'])
                start = row['start']
                end = row['end'] + 1
                seqid = str(row['seqid'])
                is_reverse = row['reverse']
                is_complement = row['complement']
                is_extended = row['is_extended']
                is_5p_extended = row['is_5p_extended']
                is_3p_extended = row['is_3p_extended']
                if is_reverse and is_3p_extended:
                    start+=1
                if is_complement and is_3p_extended:
                    end-=1



                is_complete = row['is_complete']
                fivep_trunc = row['is_5prime_boundary']
                threep_trunc = row['is_3prime_boundary']
                is_trunc = False
                if fivep_trunc or threep_trunc:
                    is_trunc = True        
                if seqid in seq_data:
                    seq = seq_data[seqid]['seq'][start:end]
                    if is_reverse and not is_complement:
                        seq = seq[::-1].translate(NT_SUB)
                    start_codon = seq[0:3]
                    stop_codon = seq[-3:]
                    if stop_codon in STOP_CODONS:
                        is_stop_valid = True
                    else:
                        is_stop_valid = False

                    if start_codon in START_CODONS:
                        is_start_valid = True
                    else:
                        is_start_valid = False

                    if is_start_valid and is_stop_valid:
                        cds_valid = True
                    else:
                        cds_valid = False


                    seqs.append({ 'id':str(id),'seqid':seqid,
                                 'locus_name':locus_name,'query_id':query_id,'qlen':row['qlen'],
                                 'start':start, 'end':end,
                                 'sub_start':row['sub_start'],
                                 'sub_ent':row['sub_end'],
                                 'ident':row['ident'],
                                 'qcovs':row['qcovs'],
                                 'bitscore':row['bitscore'],
                                 'reverse':is_reverse,
                                 'complement':is_complement,
                                 'is_complete':is_complete,
                                 'is_trunc':is_trunc,
                                 'fivep_trunc':fivep_trunc,
                                 'threep_trunc':threep_trunc,
                                 'is_extended':is_extended,
                                 'is_5p_extended':is_5p_extended,
                                 'is_3p_extended':is_3p_extended,
                                 'seq':seq,'start_codon':start_codon,'stop_codon':stop_codon,
                                 'is_stop_valid':is_stop_valid,'is_start_valid':is_start_valid,
                                 'is_cds_valid':cds_valid})
                    id+=1
        return seqs

    def extend(self,df,seqid_col, queryid_col, qstart_col, qend_col, sstart_col,send_col,slen_col, qlen_col, bitscore_col, overlap_threshold=1):
        sort_cols = [seqid_col, 'locus_name','ext_start', bitscore_col, 'ext_end', ]
        ascending_cols = [True, True, True, False, False]
        df = df.sort_values(sort_cols, ascending=ascending_cols).reset_index(drop=True)
        df = self.recursive_filter_overlap_records_dual_key(df, 'locus_name','sseqid', bitscore_col, sort_cols, ascending_cols, overlap_threshold)
        df = df.sort_values(sort_cols, ascending=ascending_cols).reset_index(drop=True)
        queries = df[queryid_col].to_list()

        #Remove incomplete hits when complete ones are present
        filtered = []
        for query in queries:
            subset = df[df[queryid_col] == query]
            complete = subset[subset['is_complete'] == True]
            num_complete = len(complete)
            if num_complete > 0:
                filtered.append(complete)
            else:
                filtered.append(subset)
        df = pd.concat(filtered, ignore_index=True)
        is_extended = []
        five_p_ext = []
        three_p_ext = []
        #Extend truncated sequences
        for idx, row in df.iterrows():
            e = False
            fivep_e = False
            threep_e = False
            if row['is_complete']:
                is_extended.append(e)
                five_p_ext.append(fivep_e)
                three_p_ext.append(threep_e)
                continue
            id = row['qseqid']
            qstart = int(row[qstart_col])
            qend = int(row[qend_col])
            qlen = int(row[qlen_col])
            sstart = int(row[sstart_col])
            send = int(row[send_col])
            slen = int(row[slen_col])
            is_rev = row['reverse']

            five_p_complete = row['is_5prime_complete']
            five_p_delta = qstart
            three_p_complete = row['is_3prime_complete']
            three_p_delta = qlen - qend
            if three_p_delta < 0:
                three_p_delta = 0


            if not five_p_complete:
                e = True
                fivep_e = True
                if is_rev:
                    sstart += five_p_delta
                else:
                    sstart -= five_p_delta 

            if sstart < 0:
                sstart = 0
            if send < 0:
                send = 0
            if not three_p_complete:
                e = True
                threep_e = True
                if is_rev:
                    send -= three_p_delta
                else:
                    send += three_p_delta

            if send >= slen:
                send = slen - 1

            is_extended.append(e)
            five_p_ext.append(fivep_e)
            three_p_ext.append(threep_e)
            
            if sstart < 0:
                sstart = 0
            if send < 0:
                send = 0

            if fivep_e:
                ext_len = abs(sstart - send) 
                while ext_len > qlen:
                    if is_rev:
                        sstart -= 1
                    else:
                        sstart += 1          
                    ext_len = abs(sstart - send) 

            
            if threep_e:
                ext_len = abs(sstart - send) 
                while ext_len > qlen:
                    if is_rev:
                        send += 1
                    else:
                        send -= 1   
                    ext_len = abs(sstart - send) 

            if sstart < 0:
                sstart = 0
            if send < 0:
                send = 0

            row[qstart_col] = qstart
            row[qend_col] = qend
            row[qlen_col] = qlen
            row[sstart_col] = sstart
            row[send_col] = send
            row[slen_col] = slen

            df.loc[idx] = row

        df['is_extended'] = is_extended
        df['is_5p_extended'] = five_p_ext
        df['is_3p_extended'] = three_p_ext
        return df

    def group_by_locus(self,df,seqid_col,query_col,qlen_col,extend_threshold_ratio = 0.2):
        sort_cols = ['locus_name',query_col,'ext_start', 'ext_end']
        ascending_cols = [True, True, True, True]
        df = df.sort_values(sort_cols,
                            ascending=ascending_cols).reset_index(drop=True)

        queries = df['qseqid'].to_list()

        #Remove incomplete hits when complete ones are present
        filtered = []
        for query in queries:
            subset = df[df['qseqid'] == query]
            complete = subset[subset['is_complete'] == True]
            num_complete = len(complete)
            if num_complete > 0:
                filtered.append(complete)
            else:
                filtered.append(subset)
        df = pd.concat(filtered, ignore_index=True)

        sort_cols = ['locus_name',seqid_col,'ext_start', 'ext_end']
        ascending_cols = [True, True, True, True]
        df = df.sort_values(sort_cols,
                            ascending=ascending_cols).reset_index(drop=True)

        loci = {}
        for idx, row in df.iterrows():
            query_id = row[query_col]
            locus_name = row['locus_name']
            ident = row['pident']
            cov = row['qcovs']
            sub_start = row['sstart']
            sub_end = row['send']
            start = row['ext_start']
            end = row['ext_end']
            qlen = row[qlen_col]
            seqid = row[seqid_col]
            is_extended = row['is_extended']
            is_reverse = row['reverse']
            is_complement = row['complement']
            is_complete = row['is_complete']
            is_5prime_boundary = row['is_5prime_boundary']
            is_3prime_boundary = row['is_3prime_boundary']
            is_5p_extended = row['is_5p_extended']
            is_3p_extended = row['is_3p_extended']
            #if not is_complement and is_reverse:
            #    start-=1

            if locus_name in loci:
                found = False
                for i in range(0,len(loci[locus_name])):
                    s = loci[locus_name][i]['start']
                    e = loci[locus_name][i]['end']
                    sid = loci[locus_name][i]['seqid']
                    if sid != seqid:
                        continue
                    if (s >= start and s <= end) or (e >= start and e<= end):
                        found = True
                        if end > e:
                            loci[locus_name][i]['end'] = end
                            break
                    if s >= end:
                        gap = s - end
                        max_gap = qlen * extend_threshold_ratio
                        if gap <= max_gap:
                            loci[locus_name][i]['end'] = end
                            found = True
                            break
                if found:
                    continue
            else:
                loci[locus_name] = []

            loci[locus_name].append({
                'seqid':seqid,
                'query_id':query_id,
                'qlen':row['qlen'],
                'bitscore':row['bitscore'],
                'start':start,
                'end':end,
                'sub_start':sub_start,
                'sub_end':sub_end,
                'ident':ident,
                'qcovs':cov,
                'reverse':is_reverse,
                'complement':is_complement,
                'is_complete':is_complete,
                'is_extended':is_extended,
                'is_5p_extended':is_5p_extended,
                'is_3p_extended': is_3p_extended,
                'is_5prime_boundary':is_5prime_boundary,
                'is_3prime_boundary':is_3prime_boundary

            })
        return loci






import json
import os
import sys
from argparse import ArgumentParser
from copy import deepcopy
from datetime import datetime
from dataclasses import dataclass, field, fields
import pandas as pd
import logging
import errno
from typing import Any
from locidex.classes.seq_intake import seq_intake
from locidex.constants import START_CODONS, STOP_CODONS, DBConfig
from locidex.utils import calc_md5
from locidex.version import __version__



logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.DEBUG)


@dataclass
class Parameters:
    mode: str 
    min_match_ident: str
    min_match_cov: str 
    max_ambiguous: str 
    max_internal_stops: str

@dataclass
class Data:
    sample_name: str
    profile: dict
    seq_data: dict
    metrics: dict = field(default_factory=dict)

    def __getitem__(self, name: str) -> Any:
        return getattr(self, str(name))
    
    def __setitem__(self, key: str, value: str) -> None:
        setattr(self, key, value)

@dataclass
class ReportData:
    db_info: DBConfig
    parameters: Parameters
    data: Data

    def __getitem__(self, name: str) -> Any:
        return getattr(self, str(name))
    
    def __setitem__(self, key: str, value: str) -> None:
        setattr(self, key, value)

    @classmethod
    def fields(cls):
        return fields(cls)

    @classmethod
    def deseriealize(cls, input: dict):
        """
        Return a ReportData object from deserialized json data
        """
        return cls(db_info=DBConfig(**input["db_info"]), 
            parameters=Parameters(**input["parameters"]), 
            data=Data(**input["data"]))



def add_args(parser=None):

    if parser is None:
        parser = ArgumentParser(
            description="Locidex Report: Generate a report from search results")
    parser.add_argument('-i','--input', type=str, required=True,help='Input seq_store file to report')
    parser.add_argument('--fasta', type=str, required=False,help='Optional: Query fasta file used to generate search results')
    parser.add_argument('-c', '--config', type=str, required=False, help='Locidex parameter config file (json)')
    parser.add_argument('-o', '--outdir', type=str, required=True, help='Output file to put results')
    parser.add_argument('-n', '--name', type=str, required=False, help='Sample name to include default=filename')
    parser.add_argument('-m', '--mode', type=str, required=False, help='Allele profile assignment [normal,conservative,fuzzy]',default='normal')
    parser.add_argument('-p', '--prop', type=str, required=False, help='Metadata label to use for aggregation',default='locus_name')
    parser.add_argument('-a', '--max_ambig', type=int, required=False, help='Maximum number of ambiguous characters allowed in a sequence',default=0)
    parser.add_argument('-s', '--max_stop', type=int, required=False, help='Maximum number of internal stop codons allowed in a sequence',default=0)
    parser.add_argument('-r', '--match_ident', type=float, required=False, 
                        help='Report match if percent identity is >= this value',default=0)
    parser.add_argument('-l','--match_cov', type=float, required=False, 
                        help='Report match if percent coverage is >= this value',default=0)
    parser.add_argument('-b','--min_len', type=float, required=False, 
                        help='Report match if query length is >= this value',default=0)
    parser.add_argument('-x','--max_len', type=float, required=False, 
                        help='Report match allele length is <= this value',default=100000)
    parser.add_argument('--override', required=False, help='Overwrite individual loci thresholds for filtering and use the global parameters',
                        action='store_true')
    parser.add_argument('--translation_table', type=int, required=False,
                        help='output directory', default=11)
    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)
    parser.add_argument('-f', '--force', required=False, help='Overwrite existing directory',
                        action='store_true')
    return parser



class seq_reporter:


    def __init__(self,data_dict,method='nucleotide',mode='normal',label='locus_name',filters=dict(),max_ambig=0,max_int_stop=0,match_ident=0,override=True):
        self.init_metrics()
        self.filters = filters
        self.override = override
        self.data_dict = {}
        self.profile = {}
        self.loci = {}
        self.db_seq_info = {}
        self.failed_seqids = set()
        self.max_ambig_count = max_ambig
        self.max_int_stop_count = max_int_stop
        self.label = label
        self.method = method
        self.mode = mode
        self.match_ident = match_ident
        self.data_dict = data_dict
        self.db_seq_info = self.data_dict["db_seq_info"]
        self.query_seq_data = self.data_dict["query_data"]['query_seq_data']
        self.query_hits = self.data_dict["query_data"]["query_hits"]
        self.locus_profile = self.data_dict['query_data']["locus_profile"]
        self.blast_columns = self.data_dict["query_hit_columns"]
        self.filter_hits()
        self.build_profile()
        

    def init_metrics(self):
        self.query_metrics = {
            'loci': {
            'no_blast_hit': 0,
            'single_blast_hit': 0,
            'multiple_blast_hits':0,
            'no_nucleotide_blast_hits':0,
            'no_protein_blast_hits':0,

        },
            'queries':{}
        }
        for dbtype in ['nucleotide','protein']:
            self.query_metrics['queries'][dbtype] = {  
                        'no_blast_hit':0,               
                        'single_blast_hit': 0,
                        'multiple_blast_hits':0,
                        'missing_start_codon':0,
                        'internal_stop_codon':0,
                        'missing_stop_codon':0,
                        'count_total_queries':0, 
                        'count_filtered_blast_queries':0,
                        'count_total_filtered_queries':0,              
                        'filtered':list()
                    }


    def filter_queries(self):
        failed_seqids = set()
        for seq_id in self.query_seq_data:
            ambig_count = int(self.query_seq_data[seq_id]['dna_ambig_count'])
            if isinstance(self.query_seq_data[seq_id]['count_internal_stop'], int):
                stop_count = int(self.query_seq_data[seq_id]['count_internal_stop'])
            else:
                stop_count = 0
            if ambig_count > self.max_ambig_count or stop_count > self.max_int_stop_count:
                failed_seqids.add(seq_id)

            if self.mode == 'conservative':
                count_internal_stop = self.query_seq_data[seq_id]['count_internal_stop']
                start_codon = self.query_seq_data[seq_id]["start_codon"]
                stop_codon = self.query_seq_data[seq_id]["end_codon"]
                fail = False
                if start_codon not in START_CODONS:
                    fail = True
                    self.query_metrics['queries']['nucleotide']['missing_start_codon']+=1
                    self.query_metrics['queries']['protein']['missing_start_codon']+=1
                
                if stop_codon not in STOP_CODONS:
                    fail = True
                    self.query_metrics['queries']['nucleotide']['missing_stop_codon']+=1
                    self.query_metrics['queries']['protein']['missing_stop_codon']+=1

                if count_internal_stop > 0:
                    fail = True
                    self.query_metrics['queries']['nucleotide']['internal_stop_codon']+=1
                    self.query_metrics['queries']['protein']['internal_stop_codon']+=1

                if fail:
                    self.query_metrics['queries']['nucleotide']['filtered'].append(seq_id)
                    self.query_metrics['queries']['protein']['filtered'].append(seq_id)
                    failed_seqids.add(seq_id)
        self.query_metrics['queries']['nucleotide']['count_total_filtered_queries'] = len(self.query_metrics['queries']['nucleotide']['filtered'])
        self.query_metrics['queries']['protein']['count_total_filtered_queries'] = len(self.query_metrics['queries']['protein']['filtered'])
        self.query_metrics['queries']['nucleotide']['count_total_queries'] = len(self.query_seq_data)
        self.query_metrics['queries']['protein']['count_total_queries'] = self.query_metrics['queries']['nucleotide']['count_total_queries']
        self.failed_seqids =  failed_seqids


    def build_profile(self):
        for lid in self.data_dict["db_seq_info"]:
            locus_name = self.db_seq_info[lid]["locus_name"]
            self.loci[lid] = locus_name
            self.profile[locus_name] = []


    def filter_hits(self):
        failed = set()
        for qid in self.query_hits:
            for dbtype in self.query_hits[qid]:
                filt = []
                if len(self.query_hits[qid][dbtype]) == 0:
                    self.query_metrics['queries'][dbtype]['no_blast_hit']+=1
                elif len(self.query_hits[qid][dbtype]) == 1:
                    self.query_metrics['queries'][dbtype]['single_blast_hit']+=1
                else:
                    self.query_metrics['queries'][dbtype]['multiple_blast_hits']+=1


                for hit in self.query_hits[qid][dbtype]:
                    hit_id = str(hit['sseqid'])
                    qlen = hit['qlen']
                    pident = hit['pident']
                    qcovs = hit['qcovs']
                    hinfo = self.db_seq_info[hit_id]

                    if self.override:
                        min_len = self.filters["dna_min_len"]
                        min_cov = self.filters["min_dna_match_cov"]
                        min_ident = self.filters["dna_min_ident"]
                        max_len = self.filters["dna_max_len"]
                    else:
                        if dbtype == 'nucleotide':
                            if "dna_min_len" not in hinfo:
                                min_len = self.filters["dna_min_len"]
                            else:
                                min_len = hinfo["dna_min_len"]
                            if "dna_max_len" not in hinfo:
                                max_len = self.filters["dna_min_len"]
                            else:
                                max_len = hinfo["dna_max_len"]
                            if "min_dna_match_cov" not in hinfo:
                                min_cov = self.filters["min_dna_match_cov"]
                            else:
                                min_cov = hinfo["min_dna_match_cov"]
                            if "dna_min_ident" not in hinfo:
                                min_ident = self.filters["dna_min_ident"]
                            else:
                                min_ident = hinfo["dna_min_ident"]
                        else:
                            if qlen < hinfo["aa_min_len"] or qlen > hinfo["aa_max_len"] or pident < hinfo["aa_min_ident"]:
                                continue
                            if "aa_min_len" not in hinfo:
                                min_len = self.filters["aa_min_len"]
                            else:
                                min_len = hinfo["aa_min_len"]
                            if "aa_max_len" not in hinfo:
                                max_len = self.filters["aa_min_len"]
                            else:
                                max_len = hinfo["aa_max_len"]
                            if "min_aa_match_cov" not in hinfo:
                                min_cov = self.filters["min_aa_match_cov"]
                            else:
                                min_cov = hinfo["min_aa_match_cov"]
                            if "aa_min_ident" not in hinfo:
                                min_ident = self.filters["aa_min_ident"]
                            else:
                                min_ident = hinfo["aa_min_ident"]
                    if qlen < min_len or qlen > max_len or pident < min_ident or qcovs < min_cov:
                        failed.add(qid)
                        self.query_metrics['queries'][dbtype]['filtered'].append(qid)
                        continue
                    filt.append(hit)
                self.query_hits[qid][dbtype] = filt
                self.query_metrics['queries'][dbtype]['filtered'] = list(set(self.query_metrics['queries'][dbtype]['filtered']))
                self.query_metrics['queries'][dbtype]['count_filtered_blast_queries'] = len(self.query_metrics['queries'][dbtype]['filtered'] )
        self.failed_seqids = self.failed_seqids | failed

    def calc_query_best_hit(self):
        best_hits = {}
        for qid in self.query_hits:
            best_hits[qid] = {}
            for dbtype in self.query_hits[qid]:
                best_hits[qid][dbtype] = []
                hit_ids = []
                hit_bit = []
                for hit in self.query_hits[qid][dbtype]:
                    hit_ids.append(str(hit['sseqid']))
                    hit_bit.append(hit['bitscore'])
                if len(hit_bit) == 0:
                    continue
                max_bit = max(hit_bit)
                top_ids = []
                for idx, value in enumerate(hit_bit):
                    if value >= max_bit:
                        top_ids.append(idx)
                best_hits[qid][dbtype] = top_ids

        return best_hits

    def get_hit_locinames(self):
        hit_names = {}
        for qid in self.query_hits:
            hit_names[qid] = {}
            for dbtype in self.query_hits[qid]:
                hit_names[qid][dbtype] = set()
                for hit in self.query_hits[qid][dbtype]:
                    hit_id = str(hit['sseqid'])
                    hinfo = self.db_seq_info[hit_id]
                    hit_name = hinfo['locus_name']
                    hit_names[qid][dbtype].add(hit_name)
        return hit_names
    
    def get_loci_to_query_map(self,hit_names,dbtype):
        loci_lookup = {}
        for qid in hit_names:
            if dbtype not in hit_names[qid]:
                continue
            for l in hit_names[qid][dbtype]:
                if l not in loci_lookup:
                    loci_lookup[l] = []
                loci_lookup[l].append(qid)
        return loci_lookup

    def allele_assignment(self,dbtype):
        self.query_best_hits = self.calc_query_best_hit()

        hit_loci_names = self.get_hit_locinames()
        loci_lookup = self.get_loci_to_query_map(hit_loci_names,dbtype)
        
        for locus in loci_lookup:
            loci_lookup[locus] = list(set(loci_lookup[locus]) - self.failed_seqids)
        self.populate_profile()

        loci_names_to_assign = set(self.profile.keys())
        assigned_loci = set()

        #Fix the values of any loci where there is a single matching query or no matching queries
        for locus_name in self.profile:
            query_hashes = self.profile[locus_name].split(',')
            if locus_name not in loci_lookup:
                continue
            num_queries = len(loci_lookup[locus_name])
            if num_queries == 1 and query_hashes[0] != '-':
                assigned_loci.add(locus_name )
            elif locus_name not in loci_lookup or len(loci_lookup[locus_name]) == 0:
                assigned_loci.add(locus_name)
                self.profile[locus_name] = '-'

        loci_names_to_assign = loci_names_to_assign - assigned_loci

        profile = deepcopy(self.locus_profile)
        for locus_name in loci_names_to_assign:
            if locus_name not in loci_lookup:
                continue
            matches = loci_lookup[locus_name ]
            num_matches = len(matches)
            if num_matches <= 1:
                assigned_loci.add(locus_name)
                continue

            for qid in matches:
                if dbtype not in self.query_best_hits[qid]:
                    continue
                best_hits = self.query_best_hits[qid][dbtype]
                best_hit_names = set()
                for l in best_hits:
                    l = str(l)
                    hinfo = self.db_seq_info[l]
                    best_hit_names.add(hinfo['locus_name'])
                if len(best_hit_names) == 1:
                    continue
                if locus_name not in best_hit_names:
                    if qid in profile[locus_name]:
                        del(profile[locus_name][qid])

        self.locus_profile = profile


    def get_matching_ref_seq_info(self,qid, dbtype):
        for hit in self.query_hits[qid][dbtype]:
            hit_id = str(hit['sseqid'])
            pident = hit['pident']
            if pident < self.match_ident:
                continue
            hinfo = self.db_seq_info[hit_id]
            return hinfo
        return {}
    
    def populate_profile(self):
        categories = {
            'no_blast_hit': 0,
            'single_blast_hit': 0,
            'multiple_blast_hits':0,
            'no_nucleotide_blast_hits':0,
            'no_protein_blast_hits':0
        }
        for locus_name in self.profile:
            values = set()
            if locus_name in self.locus_profile:
                values = set(self.locus_profile[locus_name][self.method])
            allele_hashes = []
            values = values - self.failed_seqids
            if len(values) == 0:
                categories['no_blast_hit']+=1
            nt_seqids = set(self.locus_profile[locus_name]['nucleotide']) - self.failed_seqids
            if len( nt_seqids) == 0:
                categories['no_nucleotide_blast_hits']+=1
            pr_seqids = set(self.locus_profile[locus_name]['protein']) - self.failed_seqids
            if len( pr_seqids) == 0:
                categories['no_protein_blast_hits']+=1

            for seq_id in values:
                if self.method == 'nucleotide':
                    key = "dna_hash"
                elif self.method == 'protein':
                    key = "aa_hash"
                else:
                    continue
                
                hash_value = self.query_seq_data[seq_id][key]
                

                if self.mode == 'fuzzy':
                    ref_seq_hitinfo = self.get_matching_ref_seq_info(seq_id, self.method)
                    if len(ref_seq_hitinfo) > 0:
                        if self.method == 'nucleotide':
                            hash_value = ref_seq_hitinfo['dna_seq_hash']  
                        elif self.method == 'protein':
                            hash_value = ref_seq_hitinfo['aa_seq_hash'] 
                if self.mode == 'conservative':
                    if  'protein' in  self.locus_profile[locus_name] and len( self.locus_profile[locus_name]['protein']) > 0:

                        if seq_id not in self.locus_profile[locus_name]['protein'] or seq_id not in self.locus_profile[locus_name]['nucleotide']:
                            continue        

                allele_hashes.append(hash_value)

            num_alleles = len(allele_hashes)
            unique_allele_count = len(set(allele_hashes))
            if unique_allele_count > 1:
                categories['multiple_blast_hits']+=1
            elif unique_allele_count == 1:
                categories['single_blast_hit']+=1
            if unique_allele_count > 1 and self.mode == 'conservative':
                allele_hashes = ['-']
            elif num_alleles > 1 and self.mode == 'normal':
                allele_hashes = calc_md5(["".join([str(x) for x in sorted(allele_hashes)])])
            elif num_alleles == 0:
                allele_hashes = ['-']
            elif self.mode == 'fuzzy':
                allele_hashes = calc_md5(["".join([str(x) for x in sorted(allele_hashes)])])
            self.profile[locus_name] = ",".join(list(set([str(x) for x in allele_hashes])))
        self.query_metrics['loci'] = categories


    def extract_hit_data(self,dbtype):
        data = []
        for qid in self.query_hits:
            query_name = self.query_seq_data[qid][self.label]
            for hit in self.query_hits[qid][dbtype]:
                hit_id = str(hit['sseqid'])
                hinfo = self.db_seq_info[hit_id]
                record = {'query_name':query_name}
                for fieldname in self.blast_columns:
                    record[fieldname] = hit[fieldname]
                for fieldname in hinfo:
                    record[fieldname] = hinfo[fieldname]

                data.append(record)
        return pd.DataFrame.from_dict(data)


def run_report(config):
    
    analysis_parameters = config

    #Input Parameters
    input_file = config['input']
    outdir = config['outdir']
    label = config['prop']
    sample_name = config['name']
    force = config['force']
    mode = config['mode']
    fasta_file = config['fasta']
    max_ambig = config['max_ambig']
    max_int_stop = config['max_stop']
    match_ident = config['match_ident']
    match_cov = config['match_cov']
    translation_table = config['translation_table']
    override = config['override']
    min_len = config['min_len']
    max_len = config['max_len']


    run_data = dict()
    run_data['analysis_start_time'] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    run_data['parameters'] = analysis_parameters


    if os.path.isdir(outdir) and not force:
        logger.critical(f'Error {outdir} exists, if you would like to overwrite, then specify --force')
        raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST), str(outdir))

    if not os.path.isdir(outdir):
        os.makedirs(outdir, 0o755)

    seq_store_dict = {}
    with open(input_file ,'r',encoding='utf-8') as fh:
        seq_store_dict = json.load(fh)

    if len(seq_store_dict) == 0:
        logger.critical("seq_store from file: {} is empty".format(input_file))
        raise ValueError("seq_store from file: {} is empty".format(input_file))

    if sample_name is None:
        sample_name = seq_store_dict["query_data"]["sample_name"]
    filters = {
        "dna_min_len": min_len,
        "dna_min_ident": match_ident,
        "min_dna_match_cov": match_cov,
        "dna_max_len":max_len,
        "aa_min_len": min_len,
        "aa_min_ident": match_ident,
        "min_aa_match_cov": match_cov,
        "aa_max_len":max_len
    }

    #validate the ids
    seq_data = {}
    if fasta_file is not None:
        seq_info = seq_store_dict["query_data"]["query_seq_data"]
        seq_obj = seq_intake(fasta_file, 'fasta', 'CDS', translation_table, perform_annotation=False)
        if len(seq_info) != len(seq_obj.seq_data):
            logger.critical(f'Error the supplied fasta file: {fasta_file} ({len(seq_obj.seq_data)}) seq_store file: {input_file} ({len(seq_info)}) \
                   do not have the same number of sequences. These files must be matched')
            raise ValueError(f"Supplied fasta and seq_store have different numbers of sequences: {str(fasta_file)}, {str(input_file)}")

        for i in range(0,len(seq_obj.seq_data)):
            id = str(i)
            if id not in seq_info:
                logger.critical(f'Error {id} key from fasta file not in seq_store')
                raise KeyError(f'Error {id} key from fasta file not in seq_store')
            pid_1 = seq_info[id]["seq_id"]
            pid_2 = seq_obj.seq_data[i]["seq_id"]
            if pid_1 != pid_2:
                logger.critical(f'Error seq_store key for {id}: {pid_1} mismatched to input fasta {id}: {pid_2}. These files must be matched')
                raise KeyError(f'Error seq_store key for {id}: {pid_1} mismatched to input fasta {id}: {pid_2}. These files must be matched')
            seq_data[id] = seq_obj.seq_data[i]

    allele_obj = seq_reporter(seq_store_dict, method='nucleotide', mode=mode, label=label, filters=filters,max_ambig=max_ambig,max_int_stop=max_int_stop,match_ident=match_ident,override=override)



    allele_obj.filter_queries()
    allele_obj.allele_assignment('nucleotide')
    allele_obj.extract_hit_data('nucleotide').to_csv(os.path.join(outdir,"nucleotide.hits.txt"),header=True,sep="\t", index=False)
    allele_obj.extract_hit_data('protein').to_csv(os.path.join(outdir, "protein.hits.txt"), header=True, sep="\t", index=False)


    profile = ReportData(
        db_info=DBConfig(**seq_store_dict["db_info"]),
        parameters= Parameters(
            mode=mode,
            min_match_ident=match_ident,
            min_match_cov=match_cov,
            max_ambiguous=max_ambig,
            max_internal_stops=max_int_stop
        ),
        data = Data(
            sample_name = sample_name,
            profile = {sample_name: allele_obj.profile},
            seq_data=seq_data,
            metrics= allele_obj.query_metrics
        )
    )

    
    if len(profile.data.seq_data) > 0:
        # add locus information to seq_data
        look_up = {}
        for locus_name in profile.data.profile[sample_name]:
            h = profile.data.profile[sample_name][locus_name]
            if h not in look_up:
                look_up[h] = []
            look_up[h].append(locus_name)
        
        for seq_id in profile.data.seq_data:
            h = profile.data.seq_data[seq_id]['dna_hash']
            if h in look_up:
                profile.data.seq_data[seq_id]['locus_name'] = ",".join([str(x) for x in look_up[h]])
            else:
                profile.data.seq_data[seq_id]['locus_name'] = ''


    with open(os.path.join(outdir,"report.json"),"w") as out:
        json.dump(profile,out,indent=4, default=lambda o: o.__dict__)

    run_data['result_file'] = os.path.join(outdir,"report.json")
    run_data['analysis_end_time'] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    with open(os.path.join(outdir,"run.json"),'w' ) as fh:
        fh.write(json.dumps(run_data, indent=4))


def run(cmd_args=None):
    logger.info("Beginning report")
    if cmd_args is None:
        parser = add_args()
        cmd_args = parser.parse_args()
    analysis_parameters = vars(cmd_args)
    config_file = cmd_args.config

    config = {}
    if config_file is not None:
        with open(config_file) as fh:
            config = json.loads(fh.read())

    for p in analysis_parameters:
        if not p in config:
            config[p] = analysis_parameters[p]

    run_report(config)
    logger.info("Finishing report workflow.")


# call main function
if __name__ == '__main__':
    run()






"""
Test the ever growing util functions
"""

import pytest
from locidex import utils
from pathlib import Path
from locidex import manifest
from argparse import Namespace
from collections import namedtuple


def test_check_db_groups_pass(monkeypatch):
    nm_group = Namespace(db_group="Db1", db_name="test_name", db_version="1.0.0")
    analysis_params = {"db_group": "Db1", "db_name": "test_name", "db_version": "1.0.0"}

    def mockreturn(*args, **kwargs):
        ret_tup = namedtuple('stuff', ["db_path"])
        ret_val = ret_tup(True)
        return ret_val
    monkeypatch.setattr(manifest, "get_manifest_db", mockreturn)
    analysis_params = utils.check_db_groups(analysis_params, nm_group)
    assert analysis_params["db"]

def test_check_db_groups_fail():
    nm_group = Namespace(db_group="Db1", db_name="test_name", db_version="1.0.0")
    analysis_params = {"db_group": "Db1", "db_name": "test_name"}

    with pytest.raises(KeyError):
        analysis_params = utils.check_db_groups(analysis_params, nm_group)


@pytest.mark.parametrize( "file_in,type",
    [
        ("test.fa", "fasta"),
        ("test.fas", "fasta"),
        ("test.ffn", "fasta"),
        ("test.fna", "fasta"),
        ("test.fasta.gz", "fasta"),
        ("test.fas.gz", "fasta"),
        ("test.fa.gz", "fasta"),
        ("test.fna.gz", "fasta"),
        ("test.ffn.gz", "fasta"),
        ("test.gbk", "genbank"),
        ("test.genbank", "genbank"),
        ("test.gbf", "genbank"),
        ("test.gbk.gz", "genbank"),
        ("test.genbank.gz", "genbank"),
        ("test.gbf.gz", "genbank"),
        ("test.gbff.gz", "genbank"),
        ("test.gbff", "genbank"),
        ("test.1.fasta.gz", "fasta"),
    ])
def test_get_format(file_in, type):
    """
    test get_format function
    """
    assert utils.get_format(Path(file_in)) == type



from dataclasses import dataclass, asdict, fields
import pathlib
import logging
import os
import errno
from typing import Any, Union, NamedTuple, Optional

DNA_AMBIG_CHARS = ['b', 'd', 'e', 'f', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 'u', 'v', 'w', 'x',
                   'y', 'z', '-']
DNA_IUPAC_CHARS = ['b', 'd', 'e', 'f', 'h', 'i', 'j', 'k', 'l', 'm', 'o', 'p', 'q', 'r', 's', 'u', 'v', 'w', 'x', 'y',
                   'z']

NT_SUB = str.maketrans('acgtrymkswhbvdnxACGTRYMKSWHBVDNX',
                       'tgcayrkmswdvbhnxTGCAYRKMSWDVBHNX')

PROTEIN_ALPHA = ['a', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'k', 'l', 'm', 'n', 'p', 'q', 'r', 's', 't', 'v', 'w', 'y', 'x','*','b', 'z','j' ]
DNA_ALPHA = ['a', 'c', 'g', 't', 'r', 'y', 's', 'w', 'k', 'm', 'b', 'd', 'h', 'v', 'n']

START_CODONS = ['atg','gtg','ctg','ttg','ata']
STOP_CODONS = ['taa','tag','tga']


@dataclass(frozen=True)
class CharacterConstants:
    stop_codon: str = "*"


def raise_file_not_found_e(file: Union[str, pathlib.Path], logger: logging.Logger):
    """
    Raise a file not found error.
    """
    logger.critical("File supplied does not exist: {}".format(str(file)))
    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(file))


class BlastColumns(NamedTuple):
    qseqid: str
    sseqid: str
    qlen: int
    slen: int
    qstart: int
    qend: int
    sstart: int
    send: int
    length: int
    mismatch: str
    pident: float
    qcovhsp: float
    qcovs: float
    sstrand: str
    evalue: float
    bitscore: float

@dataclass(frozen=True)
class BlastCommands:
    # upgrading this to a string enum would be nice
    tblastn: str = "tblastn"
    blastn: str = "blastn"
    blastp: str = "blastp"

    @classmethod
    def _keys(cls) -> list:
        return [i.name for i in fields(cls)]

FILE_TYPES = {
    'genbank': frozenset([".gbk",".genbank",".gbf",".gbk.gz",".genbank.gz",".gbf.gz",".gbff",".gbff.gz"]),
    'fasta': frozenset([".fasta",".fas",".fa",".ffn",".fna",".fasta.gz",".fas.gz",".fa.gz",".ffn.gz",".fna.gz"]),
}


SCHEME_HEADER = [
    'uid',
    'locus_id',
    'locus_meta',
    'is_cds',
    'max_internal_stop',
    'min_dna_len',
    'max_dna_len',
    'dna_seq',
    'dna_seq_len',
    'dna_seq_hash',
    'aa_seq',
    'aa_seq_len',
    'aa_seq_hash',
    'start_codon',
    'stop_codon'
]

EXTRACT_MODES = ['snps','trim','raw','extend']

# Manifest opts for parsing


OPTION_GROUPS = {
    "db_group": ["db_name", "db_version"],
}

@dataclass
class DBConfig:
    db_name: Optional[str] = None
    db_version: Optional[str] = None
    db_date: Optional[str] = None
    db_author: Optional[str] = None
    db_desc: Optional[str] = None
    db_num_seqs: Optional[Union[str, int]] = None
    is_nucl: Optional[bool] = None
    is_prot: Optional[bool] = None
    nucleotide_db_name: Optional[str] = None
    protein_db_name: Optional[str] = None

    def __getitem__(self, name: str) -> Any:
        return getattr(self, str(name))
    
    def __setitem__(self, key: str, value: str) -> None:
        setattr(self, key, value)
    
    def to_dict(self) -> dict:
        return asdict(self)

    @classmethod
    def _keys(cls) -> list:
        return [i.name for i in fields(cls)]

    def keys(self) -> list:
        return [i.name for i in fields(self)]


@dataclass(frozen=True)
class DBFiles:
    meta_file: str = "meta.json"
    config_file: str = "config.json"
    results_file: str = "results.json"
    blast_dir: str = "blast"

    @classmethod
    def items(cls):
        return [(i.name, pathlib.Path(i.default)) for i in fields(cls)]

@dataclass(frozen=True)
class ManifestFields:
    db_path: str = "path"
    config_data: str = "config"



DB_EXPECTED_FILES = {
    'config':'config.json',
    'meta':'meta.json',
    'nucleotide':'nucleotide.fasta',
    'protein':'protein.fasta',
}

@dataclass
class MetadataFields:
    num_seqs: int
    is_cds: bool
    trans_table: int
    dna_min_len: int
    dna_max_len: int
    dna_min_ident: float
    aa_min_len: int
    aa_max_len: int
    aa_min_ident: float
    
    def to_dict(self):
        return asdict(self)



class LocidexDBHeader(NamedTuple):
    seq_id: str
    locus_name: str
    locus_name_alt: str 
    locus_product: str
    locus_description: str
    locus_uid: str
    dna_seq: str
    dna_seq_len: int
    dna_seq_hash: str
    aa_seq: Optional[str]
    aa_seq_len: Optional[int]
    aa_seq_hash: Optional[str]
    dna_min_len: int
    dna_max_len: int
    aa_min_len: Optional[int]
    aa_max_len: Optional[int]
    dna_min_ident: float
    aa_min_ident: Optional[float]
    min_dna_match_cov: int
    min_aa_match_cov: Optional[int]
    count_int_stops: int
    dna_ambig_count: int

UTILITIES_CHECK = [
    "makeblastdb",
    BlastCommands.tblastn,
    BlastCommands.blastn,
    BlastCommands.blastp,
    "mafft",
    "blastdbcheck"
]


import pytest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import gzip
from locidex.classes.prodigal import gene_prediction

# Helper function to create a FASTA file
def create_fasta_file(path, seq_records, compress=False):
    mode = 'wt'
    _open = gzip.open if compress else open
    with _open(path, mode) as f:
        SeqIO.write(seq_records, f, "fasta")

@pytest.fixture
def fasta_file(tmp_path):
    # Create a FASTA file with the 16S rRNA gene from Escherichia coli
    file_path = tmp_path / "ecoli_16S_rRNA.fasta"
    ecoli_16S_seq = Seq("AGAGTTTGATCCTGGCTCAGAACGAACGCTGGCGGCAGGCCTAACACATGCAAGTCGAACGGTAACAGGAAGA")
    seqs = [SeqRecord(ecoli_16S_seq, id="Ecoli_16S_rRNA")]
    create_fasta_file(file_path, seqs)
    return file_path

@pytest.fixture
def gzip_fasta_file(tmp_path):
    file_path = tmp_path / "ecoli_16S_rRNA_compressed.fasta.gz"
    ecoli_16S_seq = Seq("AGAGTTTGATCCTGGCTCAGAACGAACGCTGGCGGCAGGCCTAACACATGCAAGTCGAACGGTAACAGGAAGA")
    seqs = [SeqRecord(ecoli_16S_seq, id="Ecoli_16S_rRNA")]
    create_fasta_file(file_path, seqs, compress=True)
    return file_path

def test_initialization_nonexistent_file():
    non_existent_filename = "nonexistent.fasta"
    gp = gene_prediction(non_existent_filename)
    assert not gp.status, "Status should be False for non-existent file"
    assert f'Error {non_existent_filename} does not exist' in gp.messages, "Proper error message should be in messages"

def test_gene_prediction_normal_fasta(fasta_file):
    predictor = gene_prediction(str(fasta_file))
    predictor.predict()
    assert predictor.status, "Should be true if gene prediction was successful"
    assert len(predictor.sequences) > 0, "Sequences should be populated"
    assert len(predictor.genes) > 0, "Sequences should be populated"

def test_gene_prediction_gzip_fasta(gzip_fasta_file):
    predictor_gzip = gene_prediction(str(gzip_fasta_file))
    predictor_gzip.predict()
    assert predictor_gzip.status, "Should be true if gene prediction was successful"
    assert len(predictor_gzip.sequences) > 0, "Sequences should be populated"
    assert len(predictor_gzip.genes) > 0, "Sequences should be populated"


import pytest
import json
import warnings
from locidex.classes.metadata import metadata

@pytest.fixture
def good_metadata(tmp_path):
    good_meta = tmp_path / "good_metadata.json"
    good_meta.write_text(json.dumps({"key1": "value1", "key2": "value2", "key3": "value3"}))
    return good_meta

@pytest.fixture
def incomplete_metadata(tmp_path):
    incomplete_meta = tmp_path / "incomplete_metadata.json"
    incomplete_meta.write_text(json.dumps({"key1": "value1", "key2": "value2"}))
    return incomplete_meta

@pytest.mark.skip(reason="Need to handle None data gracefully. Class improvement needed.")
def test_metadata_good(good_metadata):
    required_keys = ['key1', 'key2', 'key3']
    warnings.warn("Class should handle JSON parsing failures and return appropriate errors.", UserWarning)
    meta = metadata(good_metadata, required_keys)
    assert meta.status, "Metadata should be valid with all required keys present"

@pytest.mark.skip(reason="Class should explicitly check for None before iterating over dictionary.")
def test_metadata_incomplete(incomplete_metadata):
    required_keys = ['key1', 'key2', 'key3']
    warnings.warn("Class should explicitly check for None before iterating over dictionary to handle missing keys gracefully.", UserWarning)
    meta = metadata(incomplete_metadata, required_keys)
    assert not meta.status, "Metadata should be invalid due to missing key"

@pytest.mark.skip(reason="__init__ method should handle non-existent files properly.")
def test_metadata_non_existent(tmp_path):
    non_existent_file = tmp_path / "does_not_exist.json"
    required_keys = ['key1', 'key2', 'key3']
    warnings.warn("__init__ method should handle non-existent files properly and should not return any value.", UserWarning)
    meta = metadata(str(non_existent_file), required_keys)
    assert not meta.status, "Metadata should be invalid due to non-existent file"


import json
from datetime import datetime
import pandas as pd
import os, sys
from pathlib import Path
from argparse import (ArgumentParser, ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter)
from locidex.version import __version__
from locidex.constants import DBFiles
from locidex.classes import run_command
from locidex.constants import DBConfig, MetadataFields, raise_file_not_found_e
from locidex.classes.blast import BlastMakeDB
from locidex.manifest import DBData
from locidex.utils import get_user
import errno
import logging
import sys

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.INFO)

class locidex_build:


    def __init__(self, input_file: os.PathLike, outdir: os.PathLike, config: DBConfig, 
                seq_columns={DBData.nucleotide_name():'dna_seq',DBData.protein_name():'aa_seq'},force=False,parse_seqids=False,translation_table: int = 11):
        self.input_file = input_file
        self.translation_table = translation_table
        self.outdir = outdir
        self.force = force
        self.status = self.init_dir(self.outdir)
        self.parse_seqids = parse_seqids

        self.is_dna = False
        self.is_protein = False
        self.blast_dir = outdir.joinpath("blast")

        self.init_dir(self.blast_dir)

        self.df = self.read_data(self.input_file)
        self.config = config
        self.config.db_num_seqs = len(self.df)

        for t in seq_columns:
            col_name = seq_columns[t]
            s = self.is_seqtype_present(col_name)
            if s:
                outfile  = self.blast_dir.joinpath(t)
                if t == DBData.nucleotide_name():
                    self.is_dna = True
                    self.config.nucleotide_db_name = t
                    blast_method = DBData.nucleotide_db_type()
                elif t == DBData.protein_name():
                    self.is_protein  = True
                    self.config.protein_db_name = t
                    blast_method = DBData.protein_db_type()
                self.init_dir(outfile)
                out_fasta = outfile.joinpath("{}.fasta".format(t))
                output_db_path = outfile.joinpath(t)
                self.write_seq_file(out_fasta, col_name)
                creating_db = BlastMakeDB(input_file=out_fasta, 
                                        db_type=blast_method,
                                        parse_seqids=self.parse_seqids,
                                        output_db_path=output_db_path)
                creating_db.makeblastdb()
        self.config.is_nucl = self.is_dna
        self.config.is_prot = self.is_protein
        self.get_metadata(self.df,columns_to_exclude=list(seq_columns.values()))

    def init_dir(self, d: Path):
        logger.info("Checking if directory {} exists.".format(d))
        try:
            d.mkdir(parents=True, exist_ok=self.force, mode=0o755)
        except FileExistsError:
            logger.critical("Database file {} already exists. To overwrite please run with --force".format(d))
            raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST), str(d))
        return True
    

    def read_data(self,f):
        if os.path.isfile(f) and os.path.getsize(f) > 0:
            return pd.read_csv(f,sep="\t",header=0)
        else:
            return pd.DataFrame()

    def is_seqtype_present(self,col_name):
        s = True
        if not col_name in self.df:
            return False
        values = list(self.df[col_name].dropna().astype('str').unique())
        if len(values) == 0:
            return False
        return True


    def write_seq_file(self,f,col_name):
        oh = open(f, "w")
        data = dict(zip(self.df.index.tolist(), self.df[col_name]))
        for id in data:
            oh.write(f">{id}\n{data[id]}\n")
        oh.close()

    def get_metadata(self,df,columns_to_exclude=['dna_seq','aa_seq']):
        columns_to_exclude = set(columns_to_exclude)
        columns_to_include = []
        for col in list(df.columns):
            if col not in columns_to_exclude:
                columns_to_include.append(col)
        subset = df[columns_to_include]
        self.meta = {
            "info": MetadataFields(
                num_seqs=len(df),
                is_cds=True,
                trans_table=self.translation_table,
                dna_min_len=min(df['dna_min_len'].tolist()),
                dna_max_len=max(df['dna_min_len'].tolist()),
                dna_min_ident=min(df['dna_min_ident'].tolist()),
                aa_min_len=min(df['aa_min_len'].tolist()),
                aa_max_len= max(df['aa_min_len'].tolist()),
                aa_min_ident= min(df['aa_min_ident'].tolist())).to_dict(),
            "meta": subset.to_dict(orient='index')
        }

def add_args(parser=None):
    translation_table_def = 11
    default_version = "1.0.0"
    if parser is None:
        parser = ArgumentParser(
            description="Locidex: Build a locidex database",)
    parser.add_argument('-i','--input_file', type=str, required=True,help='Input tsv formated for locidex')
    parser.add_argument('-o', '--outdir', type=str, required=True, help='Output directory to put results')
    parser.add_argument('-n', '--name', type=str, required=True, help='Database name',default='Locidex Database')
    parser.add_argument('-a', '--author', type=str, required=False, help='Author Name for Locidex Database. Default: {}'.format(get_user()),default=get_user())
    parser.add_argument('-c', '--db_ver', type=str, required=False, help='Version code for locidex db: {}'.format(default_version),
                        default=default_version)
    parser.add_argument('-e', '--db_desc',type=str, required=False, help='Version code for locidex db',
                        default='')
    parser.add_argument("-t", "--translation_table", type=int, required=False, help="Translation table to use: {}".format(translation_table_def), default=translation_table_def)
    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)
    parser.add_argument('-f', '--force', required=False, help='Overwrite existing directory',
                        action='store_true')
    return parser
    



def run(cmd_args=None):
    if cmd_args is None:
        parser = add_args()
        cmd_args = parser.parse_args()
    
    input_file = cmd_args.input_file
    outdir = cmd_args.outdir
    force = cmd_args.force


    config = DBConfig(
        db_name=cmd_args.name,
        db_version =cmd_args.db_ver,
        db_desc=cmd_args.db_desc,
        db_author=cmd_args.author,
        db_date=datetime.now().strftime("%Y-%m-%d"),
    )

    run_params = vars(cmd_args)
    run_params = run_params | config.to_dict()
    run_data = dict()
    run_data['analysis_start_time'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    run_data['parameters'] = vars(cmd_args)

    if not os.path.isfile(input_file):
        logger.critical(f'Error {input_file} does not exist, please check path and try again')
        raise_file_not_found_e(input_file, logger)

    obj = locidex_build(Path(input_file), Path(outdir),config=config,seq_columns={'nucleotide':'dna_seq','protein':'aa_seq'},force=force)

    if obj.status == False:
        logger.critical(f'Error something went wrong building the db.')
        raise ValueError("Something went wrong building db.")

    run_data['analysis_end_time'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    with open(os.path.join(outdir,DBFiles.config_file),"w") as oh:
        oh.write(json.dumps(obj.config.to_dict(),indent=4))

    with open(os.path.join(outdir, DBFiles.meta_file),"w") as oh:
        oh.write(json.dumps(obj.meta,indent=4))


    with open(os.path.join(outdir, DBFiles.results_file),"w") as oh:
        oh.write(json.dumps(run_data,indent=4))


# call main function
if __name__ == '__main__':
    run()






import json
import os
import re
import shutil
import sys
from pathlib import Path
from argparse import (ArgumentParser, ArgumentDefaultsHelpFormatter, RawDescriptionHelpFormatter)
from datetime import datetime
import numpy as np
import pandas as pd
import logging
import errno
from locidex.classes.extractor import extractor
from locidex.classes.blast import BlastSearch, FilterOptions, BlastMakeDB
from locidex.manifest import DBData
from locidex.classes.db import search_db_conf, db_config
from locidex.classes.seq_intake import seq_intake, seq_store
from locidex.constants import FILE_TYPES, BlastColumns, BlastCommands, DBConfig, DB_EXPECTED_FILES, EXTRACT_MODES, raise_file_not_found_e, START_CODONS, STOP_CODONS
from locidex.version import __version__
from locidex.classes.aligner import perform_alignment, aligner
from locidex.utils import check_db_groups, get_format,translate_dna

logger = logging.getLogger(__name__)
logging.basicConfig(filemode=sys.stderr, level=logging.INFO)

def add_args(parser=None):
    if parser is None:
        parser = ArgumentParser(
            description="Locidex: Extract",)

    parser.add_argument('-i','--in_fasta', type=str, required=True,help='Query assembly sequence file (fasta)')
    parser.add_argument('-o', '--outdir', type=str, required=True, help='Output directory to put results')
    parser.add_argument('-n', '--name', type=str, required=False, help='Sample name to include default=filename')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-d', '--db', type=str, required=False, help='Locidex database directory')
    group.add_argument("--db_group", type=str, required=False, help="A directory of databases containing a manifest file. Requires the db_name option to be set to select the correct db")
    parser.add_argument('--db_name', type=str, required=False, help='Name of database to perform search, used when a manifest is specified as a db')
    parser.add_argument('--db_version', type=str, required=False, help='Version of database to perform search, used when a manifest is specified as a db')
    parser.add_argument('-c', '--config', type=str, required=False, help='Locidex parameter config file (json)')
    parser.add_argument('--min_evalue', type=float, required=False, help='Minumum evalue required for match',
                        default=0.0001)
    parser.add_argument('--min_dna_len', type=int, required=False, help='Global minumum query length dna',
                        default=1)
    parser.add_argument('--min_aa_len', type=int, required=False, help='Global minumum query length aa',
                        default=1)
    parser.add_argument('--max_dna_len', type=int, required=False, help='Global maximum query length dna',
                        default=10000000)
    parser.add_argument('--min_dna_ident', type=float, required=False, help='Global minumum DNA percent identity required for match',
                        default=80.0)
    parser.add_argument('--min_aa_ident', type=float, required=False, help='Global minumum AA percent identity required for match',
                        default=80.0)
    parser.add_argument('--min_dna_match_cov', type=float, required=False, help='Global minumum DNA percent hit coverage identity required for match',
                        default=80.0)
    parser.add_argument('--min_aa_match_cov', type=float, required=False, help='Global minumum AA percent hit coverage identity required for match',
                        default=80.0)
    parser.add_argument('--max_target_seqs', type=int, required=False, help='Maximum number of hit seqs per query',
                        default=10)
    parser.add_argument('--keep_truncated', required=False, help='Keep sequences where match is broken at the end of a sequence',
                        action='store_true')
    parser.add_argument('--mode', type=str, required=False, help='Select from the options provided',
                        default='raw', choices=EXTRACT_MODES)
    parser.add_argument('--n_threads','-t', type=int, required=False,
                        help='CPU Threads to use', default=1)
    parser.add_argument('--format', type=str, required=False,
                        help='Format of query file [genbank,fasta]')
    parser.add_argument('--translation_table', type=int, required=False,
                        help='output directory', default=11)
    parser.add_argument('--protein_coding', type=bool, required=False,
                        help='output directory', default=True)
    parser.add_argument('-V', '--version', action='version', version="%(prog)s " + __version__)
    parser.add_argument('-f', '--force', required=False, help='Overwrite existing directory',
                        action='store_true')
    return parser

def write_seq_info(seq_data,out_file):
    data = {}
    for idx,entry in enumerate(seq_data):
        data[idx] = entry
    pd.DataFrame.from_dict(data,orient='index').to_csv(out_file,sep="\t",header=True,index=False)



def run_extract(config):
    # Input Parameters
    input_fasta = Path(config['in_fasta'])
    outdir = Path(config['outdir'])
    db_dir = config['db']
    min_dna_ident = config['min_dna_ident']
    min_evalue = config['min_evalue']
    min_dna_match_cov = config['min_dna_match_cov']
    n_threads = config['n_threads']
    translation_table = config['translation_table']
    force = config['force']
    min_dna_len = config['min_dna_len']
    keep_truncated = config['keep_truncated']
    sample_name = config['name']
    max_target_seqs = config['max_target_seqs']
    mode = config['mode'].lower()
    db_data = DBData(db_dir=db_dir)
    protein_coding = config['protein_coding']

    if not mode in EXTRACT_MODES:
        logger.critical('Provided mode for allele extraction is not valid: {}, needs to be one of ({})'.format(mode, ", ".join(EXTRACT_MODES)))
        raise ValueError('Extraction  mode is not valid: {}, needs to be one of ({})'.format(mode,EXTRACT_MODES))

    if sample_name == None:
        sample_name = os.path.basename(input_fasta)

    config = config | db_data.config_data.to_dict() # update config data to use db data
    run_data = dict()
    run_data['analysis_start_time'] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    run_data['parameters'] = config

    format = None
    if format is None:
        format = get_format(input_fasta)
    else:
        format = format.lower()

    if format is None or format not in FILE_TYPES:
        if format is None:
            logger.critical(f'Could not guess format for {input_fasta}')
            raise ValueError("Could not guess input type for: {}".format(str(input_fasta)))
        else:
            logger.critical(f'Format for query file must be one of {list(FILE_TYPES.keys())}, you supplied {format}')
            raise ValueError(f'Format for query file must be one of {list(FILE_TYPES.keys())}, you supplied {format}')

    seq_obj = seq_intake(input_fasta, format, 'source', translation_table, perform_annotation=False,skip_trans=True)

    # Validate database is valid
    db_database_config = search_db_conf(db_dir, DB_EXPECTED_FILES, DBConfig._keys())
    if db_database_config.status == False:
        logger.critical(f'There is an issue with provided db directory: {db_dir}\n {db_database_config.messages}')
        raise ValueError("There is an issue with the provided database: {}".format(db_dir))

    metadata_path = db_database_config.meta_file_path
    metadata_obj = db_config(metadata_path, ['meta', 'info'])

    if os.path.isdir(outdir) and not force:
        logger.critical(f'Error {outdir} exists, if you would like to overwrite, then specify --force')
        raise IsADirectoryError(errno.EISDIR, os.strerror(errno.EISDIR), str(outdir))

    db_path = os.path.join(outdir, 'blast_db')

    if not os.path.isdir(outdir):
        os.makedirs(outdir, 0o755)
    else:
        if os.path.isdir(db_path):
            shutil.rmtree(db_path)


    if not os.path.isdir(db_path):
        os.makedirs(db_path, 0o755)

    contigs_path = os.path.join(db_path,'contigs.fasta')
    seq_data = {}
    with open(contigs_path,'w') as oh:
        for idx,seq in enumerate(seq_obj.seq_data):
            seq_data[str(idx)] = {'id':str(seq.seq_id),'seq':seq.dna_seq}
            oh.write(">{}\n{}\n".format(idx,seq.dna_seq))
    del(seq_obj)
    #TODO this should probably work on more than just nucleotides
    contigs_db = BlastMakeDB(contigs_path, DBData.nucleotide_db_type(), True, contigs_path)
    contigs_db.makeblastdb()


    blast_dir_base = os.path.join(outdir, 'blast')
    if not os.path.isdir(blast_dir_base):
        os.makedirs(blast_dir_base, 0o755)

    #blast_database_paths = db_database_config.blast_paths

    blast_params = {
        'evalue': min_evalue,
        'max_target_seqs': max_target_seqs,
        'num_threads': n_threads,
        'word_size':11
    }

    nt_db = Path("{}.fasta".format(db_data.nucleotide_blast_db))
    if not nt_db.exists():
        logger.critical("Could not find file: {}".format(nt_db))
        raise_file_not_found_e(nt_db, logger)

    filter_options = {
        'evalue':  FilterOptions(min=None, max=min_evalue, include=None),
        'pident':  FilterOptions(min=min_dna_ident, max=None, include=None),
        'qcovs':   FilterOptions(min=min_dna_match_cov, max=None, include=None),
        'qcovhsp': FilterOptions(min=min_dna_match_cov, max=None, include=None),
    }

    hit_file = os.path.join(blast_dir_base, "hsps.txt")
    # TODO is this supposed to support nucleotide and amino acid?

    obj = BlastSearch(db_data=contigs_db.output_db_path, 
                    query_path=nt_db, 
                    blast_params=blast_params, 
                    blast_method=BlastCommands.blastn,
                    blast_columns=BlastColumns._fields,
                    filter_options=filter_options)
    hit_df = obj.get_blast_data(contigs_db.output_db_path, Path(hit_file))

    hit_df['sseqid'] = hit_df['sseqid'].astype(str)
    hit_df['qseqid'] = hit_df['qseqid'].astype(str)

    #add locus data into df
    loci = []
    for idx,row in hit_df.iterrows():
        qid = str(row['qseqid'])
        loci.append(metadata_obj.config['meta'][qid]['locus_name'])

    hit_df['locus_name'] = loci

    filt_trunc = True
    if keep_truncated:
        filt_trunc = False


    exobj = extractor(hit_df,seq_data,sseqid_col='sseqid',queryid_col='qseqid',qstart_col='qstart',qend_col='qend',
                      qlen_col='qlen',sstart_col='sstart',send_col='send',slen_col='slen',sstrand_col='sstrand',
                      bitscore_col='bitscore',filter_contig_breaks=filt_trunc)

    write_seq_info(exobj.seqs,os.path.join(outdir,'seq_data.txt'))
    exobj.df.to_csv(os.path.join(outdir,'filtered.hsps.txt'),header=True,sep="\t",index=False)

    nt_db_seq_obj = seq_intake(nt_db, 'fasta', 'source', translation_table, perform_annotation=False, skip_trans=True)
    nt_db_seq_data = {}
    for idx, seq in enumerate(nt_db_seq_obj.seq_data):
        nt_db_seq_data[str(seq.seq_id)] =  seq.dna_seq

    del(nt_db_seq_obj)

    ext_seq_data = {}
    with open(os.path.join(outdir,'raw.extracted.seqs.fasta'), 'w') as oh:
        align_dir = os.path.join(outdir, 'fastas')
        if not os.path.isdir(align_dir):
            os.makedirs(align_dir, 0o755)
        aln_obj = aligner(trim_fwd=True,trim_rev=True,ext_fwd=False, ext_rev=False,fill=False, snps_only=False)
        for idx,record in enumerate(exobj.seqs):
            seq = record['seq']
            seq = re.sub('^N+','',seq)
            seq = re.sub('N+$','',seq)    
            record['seq'] = seq
            if min_dna_len > len(record['seq']):
                continue
            seq_id = "{}:{}:{}:{}".format(record['locus_name'],record['query_id'],record['seqid'],record['id'])
    
            #fix extension issues with gaps
            
            
            if record['is_5p_extended'] or record['is_3p_extended']:
                tmp_seq_data = {seq_id: {'locus_name':record['locus_name'],
                                'ref_id':record['query_id'],
                                'ref_seq':nt_db_seq_data[record['query_id']],
                                'ext_seq':seq}}
                perform_alignment(tmp_seq_data, align_dir, n_threads)
                aln_record = tmp_seq_data[seq_id]
                if not 'alignment' in aln_record:
                    continue
                ref_id = aln_record['ref_id']
                alignment = aln_record['alignment']
                if seq_id in alignment and ref_id in alignment:
                    seq = aln_obj.call_seq(seq_id,alignment[ref_id],alignment[seq_id])['seq']            

            if protein_coding:
                first_stop_pos = len(seq)
                ref_seq = nt_db_seq_data[record['query_id']]
                start_codon = ref_seq[0:3]
                stop_codon = ref_seq[-3:]

                if start_codon in START_CODONS and stop_codon in STOP_CODONS:
                    aa_seq = translate_dna(seq.replace('-',''),translation_table)
                    count_int_stop = aa_seq[:-1].count("*")
                    if count_int_stop > 0:
                        first_stop_pos = (aa_seq.find('*') + 1) * 3
                        seq = seq.replace('-','')[0:first_stop_pos]
                
            ext_seq_data[seq_id] = {'locus_name':record['locus_name'],
                                'ref_id':record['query_id'],
                                'ref_seq':nt_db_seq_data[record['query_id']],
                                'ext_seq':seq}
            oh.write(">{}\n{}\n".format(seq_id,seq.replace('-','')))
        shutil.rmtree(align_dir)


    if mode == 'trim':
        aln_obj = aligner(trim_fwd=True,trim_rev=True,ext_fwd=False, ext_rev=False,fill=False, snps_only=False)
    elif mode == 'snps':
        aln_obj = aligner(trim_fwd=True, trim_rev=True, ext_fwd=True, ext_rev=True, fill=True, snps_only=True)
    elif mode == 'extend':
        aln_obj = aligner(trim_fwd=True, trim_rev=True, ext_fwd=True, ext_rev=True, fill=False, snps_only=False)

    if mode != 'raw':
        align_dir = os.path.join(outdir, 'fastas')
        if not os.path.isdir(align_dir):
            os.makedirs(align_dir, 0o755)
        perform_alignment(ext_seq_data, align_dir, n_threads)

        with open(os.path.join(outdir, 'processed.extracted.seqs.fasta'), 'w') as oh:
            for seq_id in ext_seq_data:
                record = ext_seq_data[seq_id]
                if not 'alignment' in record:
                    continue
                
                ref_id = record['ref_id']
                alignment = record['alignment']
                if not seq_id in alignment or not ref_id in alignment:
                    continue
                variants = aln_obj.call_seq(seq_id,alignment[ref_id],alignment[seq_id])
                oh.write(">{}\n{}\n".format(seq_id, variants['seq'].replace('-','')))

        shutil.rmtree(align_dir)





def run(cmd_args=None):
    if cmd_args is None:
        parser = add_args()
        cmd_args = parser.parse_args()

    analysis_parameters = vars(cmd_args)
    analysis_parameters = check_db_groups(analysis_params=analysis_parameters, cmd_args=cmd_args)
    config_file = cmd_args.config

    config = {}
    if config_file is not None:
        with open(config_file) as fh:
            config = json.loads(fh.read())

    for p in analysis_parameters:
        if not p in config:
            config[p] = analysis_parameters[p]

    run_extract(config)


# call main function
if __name__ == '__main__':
    run()




import pytest
import os
import locidex
from pathlib import Path
from locidex.manifest import DBData
from locidex.classes import blast
from locidex.constants import BlastCommands, BlastColumns


PACKAGE_ROOT = os.path.dirname(locidex.__file__)


@pytest.fixture()
def db_data():
    db_dir = DBData(Path(PACKAGE_ROOT).joinpath("example", "build_db_mlst_out"))
    return db_dir

@pytest.fixture()
def fasta():
    return Path(PACKAGE_ROOT).joinpath("example", "search", "NC_003198.1.fasta")

def test_validate_blast_db(db_data):
    
    test_class = blast.BlastSearch(db_data, Path("home"), dict(), BlastCommands.blastn, BlastColumns._fields, dict()) 
    test_class.validate_blast_db()


def test_blast_runs(db_data, fasta, tmpdir):
    test_class = blast.BlastSearch(db_data, fasta, dict(), BlastCommands.blastn, BlastColumns._fields, dict())  
    out_file = "out.txt"
    stdout, stderr = test_class._run_blast(db_data.nucleotide_blast_db, tmpdir / out_file)
    with open(tmpdir / out_file, "r") as fp:
        assert len(fp.readlines()) == 30


def test_blast_runs(db_data, fasta, tmpdir):
    test_class = blast.BlastSearch(db_data, fasta, dict(), BlastCommands.blastn, BlastColumns._fields, dict())  
    out_file = tmpdir / "out.txt"
    bd = test_class.get_blast_data(db_data.nucleotide_blast_db, out_file)
    assert len(bd) == 30